# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

from enum import Enum
from typing import Optional

from glide.exceptions import RequestError
from glide.protobuf.command_request_pb2 import CommandRequest, SimpleRoutes
from glide.protobuf.command_request_pb2 import SlotTypes as ProtoSlotTypes


class SlotType(Enum):
    PRIMARY = 1

    REPLICA = 2
    """
    `REPLICA` overrides the `read_from_replica` configuration. If it's used the request
    will be routed to a replica, even if the strategy is `ALWAYS_FROM_MASTER`.
    """


class Route:
    def __init__(self) -> None:
        pass


# cluster routes


class AllNodes(Route):
    """
    Route request to all nodes.

    Warning:
        Don't use it with write commands, they could be routed to a replica (RO) node and fail.
    """

    pass


class AllPrimaries(Route):
    pass


class RandomNode(Route):
    """
    Route request to a random node.

    Warning:
        Don't use it with write commands, because they could be randomly routed to a replica (RO) node and fail.
    """

    pass


class SlotKeyRoute(Route):
    def __init__(self, slot_type: SlotType, slot_key: str) -> None:
        super().__init__()
        self.slot_type = slot_type
        self.slot_key = slot_key


class SlotIdRoute(Route):
    def __init__(self, slot_type: SlotType, slot_id: int) -> None:
        super().__init__()
        self.slot_type = slot_type
        self.slot_id = slot_id


class ByAddressRoute(Route):
    """Routes a request to a node by its address

    Attributes:
        host (str): The endpoint of the node. If `port` is not provided, should be in the f"{address}:{port}" format,
            where `address` is the preferred endpoint as shown in the output of the `CLUSTER SLOTS` command.
        port (Optional[int]): The port to access on the node. If port is not provided, `host` is assumed to be in
            the format f"{address}:{port}".
    """

    def __init__(self, host: str, port: Optional[int] = None) -> None:
        super().__init__()
        if port is None:
            split = host.split(":")
            if len(split) < 2:
                raise RequestError(
                    "No port provided, expected host to be formatted as {hostname}:{port}`. Received "
                    + host
                )
            self.host = split[0]
            self.port = int(split[1])
        else:
            self.host = host
            self.port = port


def to_protobuf_slot_type(slot_type: SlotType) -> ProtoSlotTypes.ValueType:
    return (
        ProtoSlotTypes.Primary
        if slot_type == SlotType.PRIMARY
        else ProtoSlotTypes.Replica
    )


def set_protobuf_route(request: CommandRequest, route: Optional[Route]) -> None:
    if route is None:
        return
    elif isinstance(route, AllNodes):
        request.route.simple_routes = SimpleRoutes.AllNodes
    elif isinstance(route, AllPrimaries):
        request.route.simple_routes = SimpleRoutes.AllPrimaries
    elif isinstance(route, RandomNode):
        request.route.simple_routes = SimpleRoutes.Random
    elif isinstance(route, SlotKeyRoute):
        request.route.slot_key_route.slot_type = to_protobuf_slot_type(route.slot_type)
        request.route.slot_key_route.slot_key = route.slot_key
    elif isinstance(route, SlotIdRoute):
        request.route.slot_id_route.slot_type = to_protobuf_slot_type(route.slot_type)
        request.route.slot_id_route.slot_id = route.slot_id
    elif isinstance(route, ByAddressRoute):
        request.route.by_address_route.host = route.host
        request.route.by_address_route.port = route.port
    else:
        raise RequestError(f"Received invalid route type: {type(route)}")



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

import subprocess
import sys
from pathlib import Path
from typing import List, Optional

from glide.config import NodeAddress

SCRIPT_FILE = (
    Path(__file__).parent.parent.parent.parent / "utils" / "cluster_manager.py"
)


class ValkeyCluster:
    def __init__(
        self,
        tls,
        cluster_mode: bool = False,
        shard_count: int = 3,
        replica_count: int = 1,
        load_module: Optional[List[str]] = None,
        addresses: Optional[List[List[str]]] = None,
    ) -> None:
        if addresses:
            self.init_from_existing_cluster(addresses)
        else:
            self.tls = tls
            args_list = [sys.executable, str(SCRIPT_FILE)]
            if tls:
                args_list.append("--tls")
            args_list.append("start")
            if cluster_mode:
                args_list.append("--cluster-mode")
            if load_module:
                if len(load_module) == 0:
                    raise ValueError(
                        "Please provide the path(s) to the module(s) you want to load."
                    )
                for module in load_module:
                    args_list.extend(["--load-module", module])
            args_list.append(f"-n {shard_count}")
            args_list.append(f"-r {replica_count}")
            p = subprocess.Popen(
                args_list,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            output, err = p.communicate(timeout=80)
            if p.returncode != 0:
                raise Exception(
                    f"Failed to create a cluster. Executed: {p}" + ":" + f"\n{err}"
                )
            self.parse_cluster_script_start_output(output)

    def parse_cluster_script_start_output(self, output: str):
        assert "CLUSTER_FOLDER" in output and "CLUSTER_NODES" in output
        lines_output = output.splitlines()
        for line in lines_output:
            if "CLUSTER_FOLDER" in line:
                splitted_line = line.split("CLUSTER_FOLDER=")
                assert len(splitted_line) == 2
                self.cluster_folder = splitted_line[1]
            if "CLUSTER_NODES" in line:
                nodes_list = []
                splitted_line = line.split("CLUSTER_NODES=")
                assert len(splitted_line) == 2
                nodes_addresses = splitted_line[1].split(",")
                assert len(nodes_addresses) > 0
                for addr in nodes_addresses:
                    host, port = addr.split(":")
                    nodes_list.append(NodeAddress(host, int(port)))
                self.nodes_addr = nodes_list

    def init_from_existing_cluster(self, addresses: List[List[str]]):
        self.cluster_folder = ""
        self.nodes_addr = []
        for [host, port] in addresses:
            self.nodes_addr.append(NodeAddress(host, int(port)))

    def __del__(self):
        if self.cluster_folder:
            args_list = [sys.executable, SCRIPT_FILE]
            if self.tls:
                args_list.append("--tls")
            args_list.extend(["stop", "--cluster-folder", self.cluster_folder])
            p = subprocess.Popen(
                args_list,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            output, err = p.communicate(timeout=20)
            if p.returncode != 0:
                raise Exception(
                    f"Failed to stop a cluster {self.cluster_folder}. Executed: {p}"
                    + ":"
                    + f"\n{err}"
                )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath("../python"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Valkey GLIDE"
copyright = "2025, Valkey GLIDE Contributors"
author = "Valkey GLIDE Contributors"
release = "1.3.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
autodoc_mock_imports = [
    "glide.glide",
    "glide.protobuf",
    "pytest",
    "google",
]  # Prevents confusion in sphinx with imports

autodoc_typehints = "description"

autodoc_default_options = {
    "ignore-module-all": True,  # Prevents duplicate warnings with methods in parent module and its submodules
}


def avoid_duplicate(app, what, name, obj, skip, options):
    # We skip some special refs and PubSub duplicate warning messages that
    # weren't captured in ignore-module-all. These PubSub attributes
    # will still appear in the documentation.
    exclusions = (
        "__weakref__",
        "__doc__",
        "__module__",
        "__dict__",
        "callback",
        "channels_and_patterns",
        "context",
        "channel",
        "message",
        "pattern",
    )
    module = getattr(obj, "__module__", "")
    formatted = f"{module}.{name}" if module else name
    if formatted in exclusions:
        return formatted
    return None


def setup(app):
    app.connect("autodoc-skip-member", avoid_duplicate)


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


html_theme = "sphinx_rtd_theme"



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

import copy
import json as OuterJson
import random
import typing
from typing import List, Optional

import pytest
from glide.async_commands.batch import ClusterBatch
from glide.async_commands.core import ConditionalChange
from glide.async_commands.server_modules import glide_json as json
from glide.async_commands.server_modules import json_batch
from glide.async_commands.server_modules.glide_json import (
    JsonArrIndexOptions,
    JsonArrPopOptions,
    JsonGetOptions,
)
from glide.config import ProtocolVersion
from glide.constants import OK
from glide.exceptions import RequestError
from glide.glide_client import GlideClusterClient, TGlideClient

from tests.test_async_client import get_random_string


def get_random_value(value_type="str"):
    if value_type == "int":
        return random.randint(1, 100)
    elif value_type == "float":
        return round(random.uniform(1, 100), 2)
    elif value_type == "str":
        return "".join(random.choices("abcdefghijklmnopqrstuvwxyz", k=5))
    elif value_type == "bool":
        return random.choice([True, False])
    elif value_type == "null":
        return None


@pytest.mark.asyncio
class TestJson:
    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_set_get(self, glide_client: TGlideClient):
        key = get_random_string(5)

        json_value = {"a": 1.0, "b": 2}
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        result = await json.get(glide_client, key, ".")
        assert isinstance(result, bytes)
        assert OuterJson.loads(result) == json_value

        result = await json.get(glide_client, key, ["$.a", "$.b"])
        assert isinstance(result, bytes)
        assert OuterJson.loads(result) == {"$.a": [1.0], "$.b": [2]}

        assert await json.get(glide_client, "non_existing_key", "$") is None
        assert await json.get(glide_client, key, "$.d") == b"[]"

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_set_get_multiple_values(self, glide_client: TGlideClient):
        key = get_random_string(5)

        assert (
            await json.set(
                glide_client,
                key,
                "$",
                OuterJson.dumps({"a": {"c": 1, "d": 4}, "b": {"c": 2}, "c": True}),
            )
            == OK
        )

        result = await json.get(glide_client, key, "$..c")
        assert isinstance(result, bytes)
        assert OuterJson.loads(result) == [True, 1, 2]

        result = await json.get(glide_client, key, ["$..c", "$.c"])
        assert isinstance(result, bytes)
        assert OuterJson.loads(result) == {"$..c": [True, 1, 2], "$.c": [True]}

        assert await json.set(glide_client, key, "$..c", '"new_value"') == OK
        result = await json.get(glide_client, key, "$..c")
        assert isinstance(result, bytes)
        assert OuterJson.loads(result) == ["new_value"] * 3

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_set_conditional_set(self, glide_client: TGlideClient):
        key = get_random_string(5)
        value = OuterJson.dumps({"a": 1.0, "b": 2})
        assert (
            await json.set(
                glide_client,
                key,
                "$",
                value,
                ConditionalChange.ONLY_IF_EXISTS,
            )
            is None
        )
        assert (
            await json.set(
                glide_client,
                key,
                "$",
                value,
                ConditionalChange.ONLY_IF_DOES_NOT_EXIST,
            )
            == OK
        )

        assert (
            await json.set(
                glide_client,
                key,
                "$.a",
                "4.5",
                ConditionalChange.ONLY_IF_DOES_NOT_EXIST,
            )
            is None
        )

        assert await json.get(glide_client, key, ".a") == b"1.0"

        assert (
            await json.set(
                glide_client,
                key,
                "$.a",
                "4.5",
                ConditionalChange.ONLY_IF_EXISTS,
            )
            == OK
        )

        assert await json.get(glide_client, key, ".a") == b"4.5"

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_get_formatting(self, glide_client: TGlideClient):
        key = get_random_string(5)
        assert (
            await json.set(
                glide_client,
                key,
                "$",
                OuterJson.dumps({"a": 1.0, "b": 2, "c": {"d": 3, "e": 4}}),
            )
            == OK
        )

        result = await json.get(
            glide_client, key, "$", JsonGetOptions(indent="  ", newline="\n", space=" ")
        )

        expected_result = b'[\n  {\n    "a": 1.0,\n    "b": 2,\n    "c": {\n      "d": 3,\n      "e": 4\n    }\n  }\n]'
        assert result == expected_result

        result = await json.get(
            glide_client, key, "$", JsonGetOptions(indent="~", newline="\n", space="*")
        )

        expected_result = b'[\n~{\n~~"a":*1.0,\n~~"b":*2,\n~~"c":*{\n~~~"d":*3,\n~~~"e":*4\n~~}\n~}\n]'
        assert result == expected_result

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_mget(self, glide_client: TGlideClient):
        key1 = get_random_string(5)
        key2 = get_random_string(5)

        json1_value = {"a": 1.0, "b": {"a": 1, "b": 2.5, "c": True}}
        json2_value = {"a": 3.0, "b": {"a": 1, "b": 4}}

        assert (
            await json.set(glide_client, key1, "$", OuterJson.dumps(json1_value)) == OK
        )
        assert (
            await json.set(glide_client, key2, "$", OuterJson.dumps(json2_value)) == OK
        )

        # Test with root JSONPath
        result = await json.mget(
            glide_client,
            [key1, key2],
            "$",
        )
        expected_result: List[Optional[bytes]] = [
            b'[{"a":1.0,"b":{"a":1,"b":2.5,"c":true}}]',
            b'[{"a":3.0,"b":{"a":1,"b":4}}]',
        ]
        assert result == expected_result

        # Retrieves the full JSON objects from multiple keys.
        result = await json.mget(
            glide_client,
            [key1, key2],
            ".",
        )
        expected_result = [
            b'{"a":1.0,"b":{"a":1,"b":2.5,"c":true}}',
            b'{"a":3.0,"b":{"a":1,"b":4}}',
        ]
        assert result == expected_result

        result = await json.mget(
            glide_client,
            [key1, key2],
            "$.a",
        )
        expected_result = [b"[1.0]", b"[3.0]"]
        assert result == expected_result

        # Retrieves the value of the 'b' field for multiple keys.
        result = await json.mget(
            glide_client,
            [key1, key2],
            "$.b",
        )
        expected_result = [b'[{"a":1,"b":2.5,"c":true}]', b'[{"a":1,"b":4}]']
        assert result == expected_result

        # Retrieves all values of 'b' fields using recursive path for multiple keys
        result = await json.mget(
            glide_client,
            [key1, key2],
            "$..b",
        )
        expected_result = [b'[{"a":1,"b":2.5,"c":true},2.5]', b'[{"a":1,"b":4},4]']
        assert result == expected_result

        # retrieves the value of the nested 'b.b' field for multiple keys
        result = await json.mget(
            glide_client,
            [key1, key2],
            ".b.b",
        )
        expected_result = [b"2.5", b"4"]
        assert result == expected_result

        # JSONPath that exists in only one of the keys
        result = await json.mget(
            glide_client,
            [key1, key2],
            "$.b.c",
        )
        expected_result = [b"[true]", b"[]"]
        assert result == expected_result

        # Legacy path that exists in only one of the keys
        result = await json.mget(
            glide_client,
            [key1, key2],
            ".b.c",
        )
        expected_result = [b"true", None]
        assert result == expected_result

        # JSONPath doesn't exist
        result = await json.mget(
            glide_client,
            [key1, key2],
            "$non_existing_path",
        )
        expected_result = [b"[]", b"[]"]
        assert result == expected_result

        # Legacy path doesn't exist
        result = await json.mget(
            glide_client,
            [key1, key2],
            ".non_existing_path",
        )
        assert result == [None, None]

        # JSONPath one key doesn't exist
        result = await json.mget(
            glide_client,
            [key1, "{non_existing_key}"],
            "$.a",
        )
        assert result == [b"[1.0]", None]

        # Legacy path one key doesn't exist
        result = await json.mget(
            glide_client,
            [key1, "{non_existing_key}"],
            ".a",
        )
        assert result == [b"1.0", None]

        # Both keys don't exist
        result = await json.mget(
            glide_client,
            ["{non_existing_key}1", "{non_existing_key}2"],
            "$a",
        )
        assert result == [None, None]

        # Test with only one key
        result = await json.mget(
            glide_client,
            [key1],
            "$.a",
        )
        expected_result = [b"[1.0]"]
        assert result == expected_result

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_del(self, glide_client: TGlideClient):
        key = get_random_string(5)

        json_value = {"a": 1.0, "b": {"a": 1, "b": 2.5, "c": True}}
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        # Non-exiseting paths
        assert await json.delete(glide_client, key, "$..path") == 0
        assert await json.delete(glide_client, key, "..path") == 0

        assert await json.delete(glide_client, key, "$..a") == 2
        assert await json.get(glide_client, key, "$..a") == b"[]"

        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        assert await json.delete(glide_client, key, "..a") == 2
        with pytest.raises(RequestError):
            assert await json.get(glide_client, key, "..a")

        result = await json.get(glide_client, key, "$")
        assert isinstance(result, bytes)
        assert OuterJson.loads(result) == [{"b": {"b": 2.5, "c": True}}]

        assert await json.delete(glide_client, key, "$") == 1
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK
        assert await json.delete(glide_client, key, ".") == 1
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK
        assert await json.delete(glide_client, key) == 1
        assert await json.delete(glide_client, key) == 0
        assert await json.get(glide_client, key, "$") is None

        # Non-existing keys
        assert await json.delete(glide_client, "non_existing_key", "$") == 0
        assert await json.delete(glide_client, "non_existing_key", ".") == 0

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_forget(self, glide_client: TGlideClient):
        key = get_random_string(5)

        json_value = {"a": 1.0, "b": {"a": 1, "b": 2.5, "c": True}}
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        # Non-existing paths
        assert await json.forget(glide_client, key, "$..path") == 0
        assert await json.forget(glide_client, key, "..path") == 0

        assert await json.forget(glide_client, key, "$..a") == 2
        assert await json.get(glide_client, key, "$..a") == b"[]"

        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        assert await json.forget(glide_client, key, "..a") == 2
        with pytest.raises(RequestError):
            assert await json.get(glide_client, key, "..a")

        result = await json.get(glide_client, key, "$")
        assert isinstance(result, bytes)
        assert OuterJson.loads(result) == [{"b": {"b": 2.5, "c": True}}]

        assert await json.forget(glide_client, key, "$") == 1
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK
        assert await json.forget(glide_client, key, ".") == 1
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK
        assert await json.forget(glide_client, key) == 1
        assert await json.forget(glide_client, key) == 0
        assert await json.get(glide_client, key, "$") is None

        # Non-existing keys
        assert await json.forget(glide_client, "non_existing_key", "$") == 0
        assert await json.forget(glide_client, "non_existing_key", ".") == 0

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_objkeys(self, glide_client: TGlideClient):
        key = get_random_string(5)

        json_value = {"a": 1.0, "b": {"a": {"x": 1, "y": 2}, "b": 2.5, "c": True}}
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        keys = await json.objkeys(glide_client, key, "$")
        assert keys == [[b"a", b"b"]]

        keys = await json.objkeys(glide_client, key, ".")
        assert keys == [b"a", b"b"]

        keys = await json.objkeys(glide_client, key, "$..")
        assert keys == [[b"a", b"b"], [b"a", b"b", b"c"], [b"x", b"y"]]

        keys = await json.objkeys(glide_client, key, "..")
        assert keys == [b"a", b"b"]

        keys = await json.objkeys(glide_client, key, "$..b")
        assert keys == [[b"a", b"b", b"c"], []]

        keys = await json.objkeys(glide_client, key, "..b")
        assert keys == [b"a", b"b", b"c"]

        # path doesn't exist
        assert await json.objkeys(glide_client, key, "$.non_existing_path") == []
        assert await json.objkeys(glide_client, key, "non_existing_path") is None

        # Value at path isnt an object
        assert await json.objkeys(glide_client, key, "$.a") == [[]]
        with pytest.raises(RequestError):
            assert await json.objkeys(glide_client, key, ".a")

        # Non-existing key
        assert await json.objkeys(glide_client, "non_exiting_key", "$") is None
        assert await json.objkeys(glide_client, "non_exiting_key", ".") is None

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_toggle(self, glide_client: TGlideClient):
        key = get_random_string(10)
        json_value = {"bool": True, "nested": {"bool": False, "nested": {"bool": 10}}}
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        assert await json.toggle(glide_client, key, "$..bool") == [False, True, None]
        assert await json.toggle(glide_client, key, "bool") is True
        assert await json.toggle(glide_client, key, "$.not_existing") == []

        assert await json.toggle(glide_client, key, "$.nested") == [None]
        with pytest.raises(RequestError):
            assert await json.toggle(glide_client, key, "nested")

        with pytest.raises(RequestError):
            assert await json.toggle(glide_client, key, ".not_existing")

        with pytest.raises(RequestError):
            assert await json.toggle(glide_client, "non_exiting_key", "$")

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_type(self, glide_client: TGlideClient):
        key = get_random_string(10)

        json_value = {
            "key1": "value1",
            "key2": 2,
            "key3": [1, 2, 3],
            "key4": {"nested_key": {"key1": [4, 5]}},
            "key5": None,
            "key6": True,
        }
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        result = await json.type(glide_client, key, "$")
        assert result == [b"object"]

        result = await json.type(glide_client, key, "$..key1")
        assert result == [b"string", b"array"]

        result = await json.type(glide_client, key, "$.key2")
        assert result == [b"integer"]

        result = await json.type(glide_client, key, "$.key3")
        assert result == [b"array"]

        result = await json.type(glide_client, key, "$.key4")
        assert result == [b"object"]

        result = await json.type(glide_client, key, "$.key4.nested_key")
        assert result == [b"object"]

        result = await json.type(glide_client, key, "$.key5")
        assert result == [b"null"]

        result = await json.type(glide_client, key, "$.key6")
        assert result == [b"boolean"]

        # Check for non-existent path in enhanced mode $.key7
        result = await json.type(glide_client, key, "$.key7")
        assert result == []

        # Check for non-existent path within an existing key (array bound)
        result = await json.type(
            glide_client, key, "$.key3[3]"
        )  # Out of bounds for the array
        assert result == []

        # Legacy path (without $) - will return None for non-existing path
        result = await json.type(glide_client, key, "key7")
        assert result is None  # Legacy path returns None for non-existent key

        # Check for multiple path match in legacy
        result = await json.type(glide_client, key, "..key1")
        assert result == b"string"

        # Check for non-existent key with enhanced path
        result = await json.type(glide_client, "non_existent_key", "$.key1")
        assert result is None

        # Check for non-existent key with legacy path
        result = await json.type(glide_client, "non_existent_key", "key1")
        assert result is None  # Returns None for legacy path when the key doesn't exist

        # Check for all types in the JSON document using JSON Path
        result = await json.type(glide_client, key, "$[*]")
        assert result == [
            b"string",
            b"integer",
            b"array",
            b"object",
            b"null",
            b"boolean",
        ]

        # Check for all types in the JSON document using legacy path
        result = await json.type(glide_client, key, "[*]")
        assert result == b"string"  # Expecting only the first type (string for key1)

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_objlen(self, glide_client: TGlideClient):
        key = get_random_string(5)

        json_value = {"a": 1.0, "b": {"a": {"x": 1, "y": 2}, "b": 2.5, "c": True}}

        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        len = await json.objlen(glide_client, key, "$")
        assert len == [2]

        len = await json.objlen(glide_client, key, ".")
        assert len == 2

        len = await json.objlen(glide_client, key, "$..")
        assert len == [2, 3, 2]

        len = await json.objlen(glide_client, key, "..")
        assert len == 2

        len = await json.objlen(glide_client, key, "$..b")
        assert len == [3, None]

        len = await json.objlen(glide_client, key, "..b")
        assert len == 3

        len = await json.objlen(glide_client, key, "..a")
        assert len == 2

        len = await json.objlen(glide_client, key)
        assert len == 2

        # path doesn't exist
        assert await json.objlen(glide_client, key, "$.non_existing_path") == []
        with pytest.raises(RequestError):
            await json.objlen(glide_client, key, "non_existing_path")

        # Value at path isnt an object
        assert await json.objlen(glide_client, key, "$.a") == [None]
        with pytest.raises(RequestError):
            await json.objlen(glide_client, key, ".a")

        # Non-existing key
        assert await json.objlen(glide_client, "non_exiting_key", "$") is None
        assert await json.objlen(glide_client, "non_exiting_key", ".") is None

        assert await json.set(glide_client, key, "$", '{"a": 1, "b": 2, "c":3, "d":4}')
        assert await json.objlen(glide_client, key) == 4

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_arrlen(self, glide_client: TGlideClient):
        key = get_random_string(5)

        json_value = '{"a": [1, 2, 3], "b": {"a": [1, 2], "c": {"a": 42}}}'
        assert await json.set(glide_client, key, "$", json_value) == OK

        assert await json.arrlen(glide_client, key, "$.a") == [3]

        assert await json.arrlen(glide_client, key, "$..a") == [3, 2, None]

        # Legacy path retrieves the first array match at ..a
        assert await json.arrlen(glide_client, key, "..a") == 3

        # Value at path is not an array
        assert await json.arrlen(glide_client, key, "$") == [None]
        with pytest.raises(RequestError):
            assert await json.arrlen(glide_client, key, ".")

        # Path doesn't exist
        assert await json.arrlen(glide_client, key, "$.non_existing_path") == []
        with pytest.raises(RequestError):
            assert await json.arrlen(glide_client, key, "non_existing_path")

        # Non-existing key
        assert await json.arrlen(glide_client, "non_existing_key", "$.a") is None
        assert await json.arrlen(glide_client, "non_existing_key", ".a") is None

        # No path
        with pytest.raises(RequestError):
            assert await json.arrlen(glide_client, key)

        assert await json.set(glide_client, key, "$", "[1, 2, 3, 4]") == OK
        assert await json.arrlen(glide_client, key) == 4

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_clear(self, glide_client: TGlideClient):
        key = get_random_string(5)

        json_value = (
            '{"obj":{"a":1, "b":2}, "arr":[1,2,3], '
            '"str": "foo", "bool": true, "int": 42, '
            '"float": 3.14, "nullVal": null}'
        )
        assert await json.set(glide_client, key, "$", json_value) == OK

        assert await json.clear(glide_client, key, "$.*") == 6
        result = await json.get(glide_client, key, "$")
        assert (
            result
            == b'[{"obj":{},"arr":[],"str":"","bool":false,"int":0,"float":0.0,"nullVal":null}]'
        )
        assert await json.clear(glide_client, key, "$.*") == 0

        assert await json.set(glide_client, key, "$", json_value) == OK
        assert await json.clear(glide_client, key, "*") == 6

        json_value = (
            '{"a": 1, "b": {"a": [5, 6, 7], "b": {"a": true}}, '
            '"c": {"a": "value", "b": {"a": 3.5}}, "d": {"a": '
            '{"foo": "foo"}}, "nullVal": null}'
        )
        assert await json.set(glide_client, key, "$", json_value) == OK

        assert await json.clear(glide_client, key, "b.a[1:3]") == 2
        assert await json.clear(glide_client, key, "b.a[1:3]") == 0
        assert (
            await json.get(glide_client, key, "$..a")
            == b'[1,[5,0,0],true,"value",3.5,{"foo":"foo"}]'
        )
        assert await json.clear(glide_client, key, "..a") == 6
        assert await json.get(glide_client, key, "$..a") == b'[0,[],false,"",0.0,{}]'

        assert await json.clear(glide_client, key, "$..a") == 0

        # Path doesn't exists
        assert await json.clear(glide_client, key, "$.path") == 0
        assert await json.clear(glide_client, key, "path") == 0

        # Key doesn't exists
        with pytest.raises(RequestError):
            await json.clear(glide_client, "non_existing_key")

        with pytest.raises(RequestError):
            await json.clear(glide_client, "non_existing_key", "$")

        with pytest.raises(RequestError):
            await json.clear(glide_client, "non_existing_key", ".")

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_numincrby(self, glide_client: TGlideClient):
        key = get_random_string(10)

        json_value = {
            "key1": 1,
            "key2": 3.5,
            "key3": {"nested_key": {"key1": [4, 5]}},
            "key4": [1, 2, 3],
            "key5": 0,
            "key6": "hello",
            "key7": None,
            "key8": {"nested_key": {"key1": 69}},
            "key9": 1.7976931348623157e308,
        }

        # Set the initial JSON document at the key
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        # Test JSONPath
        # Increment integer value (key1) by 5
        result = await json.numincrby(glide_client, key, "$.key1", 5)
        assert result == b"[6]"  # Expect 1 + 5 = 6

        # Increment float value (key2) by 2.5
        result = await json.numincrby(glide_client, key, "$.key2", 2.5)
        assert result == b"[6]"  # Expect 3.5 + 2.5 = 6

        # Increment nested object (key3.nested_key.key1[0]) by 7
        result = await json.numincrby(glide_client, key, "$.key3.nested_key.key1[1]", 7)
        assert result == b"[12]"  # Expect 4 + 7 = 12

        # Increment array element (key4[1]) by 1
        result = await json.numincrby(glide_client, key, "$.key4[1]", 1)
        assert result == b"[3]"  # Expect 2 + 1 = 3

        # Increment zero value (key5) by 10.23 (float number)
        result = await json.numincrby(glide_client, key, "$.key5", 10.23)
        assert result == b"[10.23]"  # Expect 0 + 10.23 = 10.23

        # Increment a string value (key6) by a number
        result = await json.numincrby(glide_client, key, "$.key6", 99)
        assert result == b"[null]"  # Expect null

        # Increment a None value (key7) by a number
        result = await json.numincrby(glide_client, key, "$.key7", 51)
        assert result == b"[null]"  # Expect null

        # Check increment for all numbers in the document using JSON Path (First Null: key3 as an entire object.
        # Second Null: The path checks under key3, which is an object, for numeric values).
        result = await json.numincrby(glide_client, key, "$..*", 5)
        assert (
            result
            == b"[11,11,null,null,15.23,null,null,null,1.7976931348623157e+308,null,null,9,17,6,8,8,null,74]"
        )

        # Check for multiple path match in enhanced
        result = await json.numincrby(glide_client, key, "$..key1", 1)
        assert result == b"[12,null,75]"

        # Check for non existent path in JSONPath
        result = await json.numincrby(glide_client, key, "$.key10", 51)
        assert result == b"[]"  # Expect Empty Array

        # Check for non existent key in JSONPath
        with pytest.raises(RequestError):
            await json.numincrby(glide_client, "non_existent_key", "$.key10", 51)

        # Check for Overflow in JSONPath
        with pytest.raises(RequestError):
            await json.numincrby(glide_client, key, "$.key9", 1.7976931348623157e308)

        # Decrement integer value (key1) by 12
        result = await json.numincrby(glide_client, key, "$.key1", -12)
        assert result == b"[0]"  # Expect 12 - 12 = 0

        # Decrement integer value (key1) by 0.5
        result = await json.numincrby(glide_client, key, "$.key1", -0.5)
        assert result == b"[-0.5]"  # Expect 0 - 0.5 = -0.5

        # Check 'null' value
        result = await json.numincrby(glide_client, key, "$.key7", 5)
        assert result == b"[null]"  # Expect 'null'

        # Test Legacy Path
        # Increment float value (key1) by 5 (integer)
        result = await json.numincrby(glide_client, key, "key1", 5)
        assert result == b"4.5"  # Expect -0.5 + 5 = 4.5

        # Decrement float value (key1) by 5.5 (integer)
        result = await json.numincrby(glide_client, key, "key1", -5.5)
        assert result == b"-1"  # Expect 4.5 - 5.5 = -1

        # Increment int value (key2) by 2.5 (a float number)
        result = await json.numincrby(glide_client, key, "key2", 2.5)
        assert result == b"13.5"  # Expect 11 + 2.5 = 13.5

        # Increment nested value (key3.nested_key.key1[0]) by 7
        result = await json.numincrby(glide_client, key, "key3.nested_key.key1[0]", 7)
        assert result == b"16"  # Expect 9 + 7 = 16

        # Increment array element (key4[1]) by 1
        result = await json.numincrby(glide_client, key, "key4[1]", 1)
        assert result == b"9"  # Expect 8 + 1 = 9

        # Increment a float value (key5) by 10.2 (a float number)
        result = await json.numincrby(glide_client, key, "key5", 10.2)
        assert result == b"25.43"  # Expect 15.23 + 10.2 = 25.43

        # Check for multiple path match in legacy and assure that the result of the last updated value is returned
        result = await json.numincrby(glide_client, key, "..key1", 1)
        assert result == b"76"

        # Check if the rest of the key1 path matches were updated and not only the last value
        result = await json.get(glide_client, key, "$..key1")  # type: ignore
        assert (
            result == b"[0,[16,17],76]"
        )  # First is 0 as 0 + 0 = 0, Second doesn't change as its an array type (non-numeric), third is 76 as 0 + 76 = 0

        # Check for non existent path in legacy
        with pytest.raises(RequestError):
            await json.numincrby(glide_client, key, ".key10", 51)

        # Check for non existent key in legacy
        with pytest.raises(RequestError):
            await json.numincrby(glide_client, "non_existent_key", ".key10", 51)

        # Check for Overflow in legacy
        with pytest.raises(RequestError):
            await json.numincrby(glide_client, key, ".key9", 1.7976931348623157e308)

        # Check 'null' value
        with pytest.raises(RequestError):
            await json.numincrby(glide_client, key, ".key7", 5)

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_nummultby(self, glide_client: TGlideClient):
        key = get_random_string(10)

        json_value = {
            "key1": 1,
            "key2": 3.5,
            "key3": {"nested_key": {"key1": [4, 5]}},
            "key4": [1, 2, 3],
            "key5": 0,
            "key6": "hello",
            "key7": None,
            "key8": {"nested_key": {"key1": 69}},
            "key9": 3.5953862697246314e307,
        }

        # Set the initial JSON document at the key
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        # Test JSONPath
        # Multiply integer value (key1) by 5
        result = await json.nummultby(glide_client, key, "$.key1", 5)
        assert result == b"[5]"  # Expect 1 * 5 = 5

        # Multiply float value (key2) by 2.5
        result = await json.nummultby(glide_client, key, "$.key2", 2.5)
        assert result == b"[8.75]"  # Expect 3.5 * 2.5 = 8.75

        # Multiply nested object (key3.nested_key.key1[1]) by 7
        result = await json.nummultby(glide_client, key, "$.key3.nested_key.key1[1]", 7)
        assert result == b"[35]"  # Expect 5 * 7 = 35

        # Multiply array element (key4[1]) by 1
        result = await json.nummultby(glide_client, key, "$.key4[1]", 1)
        assert result == b"[2]"  # Expect 2 * 1 = 2

        # Multiply zero value (key5) by 10.23 (float number)
        result = await json.nummultby(glide_client, key, "$.key5", 10.23)
        assert result == b"[0]"  # Expect 0 * 10.23 = 0

        # Multiply a string value (key6) by a number
        result = await json.nummultby(glide_client, key, "$.key6", 99)
        assert result == b"[null]"  # Expect null

        # Multiply a None value (key7) by a number
        result = await json.nummultby(glide_client, key, "$.key7", 51)
        assert result == b"[null]"  # Expect null

        # Check multiplication for all numbers in the document using JSON Path
        # key1: 5 * 5 = 25
        # key2: 8.75 * 5 = 43.75
        # key3.nested_key.key1[0]: 4 * 5 = 20
        # key3.nested_key.key1[1]: 35 * 5 = 175
        # key4[0]: 1 * 5 = 5
        # key4[1]: 2 * 5 = 10
        # key4[2]: 3 * 5 = 15
        # key5: 0 * 5 = 0
        # key8.nested_key.key1: 69 * 5 = 345
        # key9: 3.5953862697246314e307 * 5 = 1.7976931348623157e308
        result = await json.nummultby(glide_client, key, "$..*", 5)
        assert (
            result
            == b"[25,43.75,null,null,0,null,null,null,1.7976931348623157e+308,null,null,20,175,5,10,15,null,345]"
        )

        # Check for multiple path matches in JSONPath
        # key1: 25 * 2 = 50
        # key8.nested_key.key1: 345 * 2 = 690
        result = await json.nummultby(glide_client, key, "$..key1", 2)
        assert result == b"[50,null,690]"  # After previous multiplications

        # Check for non-existent path in JSONPath
        result = await json.nummultby(glide_client, key, "$.key10", 51)
        assert result == b"[]"  # Expect Empty Array

        # Check for non-existent key in JSONPath
        with pytest.raises(RequestError):
            await json.nummultby(glide_client, "non_existent_key", "$.key10", 51)

        # Check for Overflow in JSONPath
        with pytest.raises(RequestError):
            await json.nummultby(glide_client, key, "$.key9", 1.7976931348623157e308)

        # Multiply integer value (key1) by -12
        result = await json.nummultby(glide_client, key, "$.key1", -12)
        assert result == b"[-600]"  # Expect 50 * -12 = -600

        # Multiply integer value (key1) by -0.5
        result = await json.nummultby(glide_client, key, "$.key1", -0.5)
        assert result == b"[300]"  # Expect -600 * -0.5 = 300

        # Test Legacy Path
        # Multiply int value (key1) by 5 (integer)
        result = await json.nummultby(glide_client, key, "key1", 5)
        assert result == b"1500"  # Expect 300 * 5 = -1500

        # Multiply int value (key1) by -5.5 (float number)
        result = await json.nummultby(glide_client, key, "key1", -5.5)
        assert result == b"-8250"  # Expect -150 * -5.5 = -8250

        # Multiply int float (key2) by 2.5 (a float number)
        result = await json.nummultby(glide_client, key, "key2", 2.5)
        assert result == b"109.375"  # Expect 43.75 * 2.5 = 109.375

        # Multiply nested value (key3.nested_key.key1[0]) by 7
        result = await json.nummultby(glide_client, key, "key3.nested_key.key1[0]", 7)
        assert result == b"140"  # Expect 20 * 7 = 140

        # Multiply array element (key4[1]) by 1
        result = await json.nummultby(glide_client, key, "key4[1]", 1)
        assert result == b"10"  # Expect 10 * 1 = 10

        # Multiply a float value (key5) by 10.2 (a float number)
        result = await json.nummultby(glide_client, key, "key5", 10.2)
        assert result == b"0"  # Expect 0 * 10.2 = 0

        # Check for multiple path matches in legacy and assure that the result of the last updated value is returned
        # last updated value is key8.nested_key.key1: 690 * 2 = 1380
        result = await json.nummultby(glide_client, key, "..key1", 2)
        assert result == b"1380"  # Expect the last updated key1 value multiplied by 2

        # Check if the rest of the key1 path matches were updated and not only the last value
        result = await json.get(glide_client, key, "$..key1")  # type: ignore
        assert result == b"[-16500,[140,175],1380]"

        # Check 'null' in legacy
        with pytest.raises(RequestError):
            await json.nummultby(glide_client, key, ".key7", 5)

        # Check for non-existent path in legacy
        with pytest.raises(RequestError):
            await json.nummultby(glide_client, key, ".key10", 51)

        # Check for non-existent key in legacy
        with pytest.raises(RequestError):
            await json.nummultby(glide_client, "non_existent_key", ".key10", 51)

        # Check for Overflow in legacy
        with pytest.raises(RequestError):
            await json.nummultby(glide_client, key, ".key9", 1.7976931348623157e308)

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_strlen(self, glide_client: TGlideClient):
        key = get_random_string(10)
        json_value = {"a": "foo", "nested": {"a": "hello"}, "nested2": {"a": 31}}
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        assert await json.strlen(glide_client, key, "$..a") == [3, 5, None]
        assert await json.strlen(glide_client, key, "a") == 3

        assert await json.strlen(glide_client, key, "$.nested") == [None]
        with pytest.raises(RequestError):
            assert await json.strlen(glide_client, key, "nested")

        with pytest.raises(RequestError):
            assert await json.strlen(glide_client, key)

        assert await json.strlen(glide_client, key, "$.non_existing_path") == []
        with pytest.raises(RequestError):
            await json.strlen(glide_client, key, ".non_existing_path")

        assert await json.strlen(glide_client, "non_exiting_key", ".") is None
        assert await json.strlen(glide_client, "non_exiting_key", "$") is None

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_strappend(self, glide_client: TGlideClient):
        key = get_random_string(10)
        json_value = {"a": "foo", "nested": {"a": "hello"}, "nested2": {"a": 31}}
        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        assert await json.strappend(glide_client, key, '"bar"', "$..a") == [6, 8, None]
        assert await json.strappend(glide_client, key, OuterJson.dumps("foo"), "a") == 9

        json_str = await json.get(glide_client, key, ".")
        assert isinstance(json_str, bytes)
        assert OuterJson.loads(json_str) == {
            "a": "foobarfoo",
            "nested": {"a": "hellobar"},
            "nested2": {"a": 31},
        }

        assert await json.strappend(
            glide_client, key, OuterJson.dumps("bar"), "$.nested"
        ) == [None]

        with pytest.raises(RequestError):
            await json.strappend(glide_client, key, OuterJson.dumps("bar"), ".nested")

        with pytest.raises(RequestError):
            await json.strappend(glide_client, key, OuterJson.dumps("bar"))

        assert (
            await json.strappend(
                glide_client, key, OuterJson.dumps("try"), "$.non_existing_path"
            )
            == []
        )
        with pytest.raises(RequestError):
            await json.strappend(
                glide_client, key, OuterJson.dumps("try"), "non_existing_path"
            )

        with pytest.raises(RequestError):
            await json.strappend(
                glide_client, "non_exiting_key", OuterJson.dumps("try")
            )

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    @typing.no_type_check  # since this is a complex test, skip typing to be more effective
    async def test_json_arrinsert(self, glide_client: TGlideClient):
        key = get_random_string(10)

        assert (
            await json.set(
                glide_client,
                key,
                "$",
                """
            {
                "a": [],
                "b": { "a": [1, 2, 3, 4] },
                "c": { "a": "not an array" },
                "d": [{ "a": ["x", "y"] }, { "a": [["foo"]] }],
                "e": [{ "a": 42 }, { "a": {} }],
                "f": { "a": [true, false, null] }
            }
            """,
            )
            == OK
        )

        # Insert different types of values into the matching paths
        result = await json.arrinsert(
            glide_client,
            key,
            "$..a",
            0,
            ['"string_value"', "123", '{"key": "value"}', "true", "null", '["bar"]'],
        )
        assert result == [6, 10, None, 8, 7, None, None, 9]

        updated_doc = await json.get(glide_client, key)

        expected_doc = {
            "a": ["string_value", 123, {"key": "value"}, True, None, ["bar"]],
            "b": {
                "a": [
                    "string_value",
                    123,
                    {"key": "value"},
                    True,
                    None,
                    ["bar"],
                    1,
                    2,
                    3,
                    4,
                ],
            },
            "c": {"a": "not an array"},
            "d": [
                {
                    "a": [
                        "string_value",
                        123,
                        {"key": "value"},
                        True,
                        None,
                        ["bar"],
                        "x",
                        "y",
                    ]
                },
                {
                    "a": [
                        "string_value",
                        123,
                        {"key": "value"},
                        True,
                        None,
                        ["bar"],
                        ["foo"],
                    ]
                },
            ],
            "e": [{"a": 42}, {"a": {}}],
            "f": {
                "a": [
                    "string_value",
                    123,
                    {"key": "value"},
                    True,
                    None,
                    ["bar"],
                    True,
                    False,
                    None,
                ]
            },
        }

        assert OuterJson.loads(updated_doc) == expected_doc

        # Insert into a specific index (non-zero)
        result = await json.arrinsert(
            glide_client,
            key,
            "$..a",
            2,
            ['"insert_at_2"'],
        )
        assert result == [7, 11, None, 9, 8, None, None, 10]

        # Check document after insertion at index 2
        updated_doc_at_2 = await json.get(glide_client, key)
        expected_doc["a"].insert(2, "insert_at_2")
        expected_doc["b"]["a"].insert(2, "insert_at_2")
        expected_doc["d"][0]["a"].insert(2, "insert_at_2")
        expected_doc["d"][1]["a"].insert(2, "insert_at_2")
        expected_doc["f"]["a"].insert(2, "insert_at_2")
        assert OuterJson.loads(updated_doc_at_2) == expected_doc

        # Insert with a legacy path
        result = await json.arrinsert(
            glide_client,
            key,
            "..a",  # legacy path
            0,
            ['"legacy_value"'],
        )
        assert (
            result == 8
        )  # Returns length of the first modified array (in this case, 'a')

        # Check document after insertion at root legacy path (all matching arrays should be updated)
        updated_doc_legacy = await json.get(glide_client, key)

        # Update `expected_doc` with the new value inserted at index 0 of all matching arrays
        expected_doc["a"].insert(0, "legacy_value")
        expected_doc["b"]["a"].insert(0, "legacy_value")
        expected_doc["d"][0]["a"].insert(0, "legacy_value")
        expected_doc["d"][1]["a"].insert(0, "legacy_value")
        expected_doc["f"]["a"].insert(0, "legacy_value")

        assert OuterJson.loads(updated_doc_legacy) == expected_doc

        # Insert with an index out of range for some arrays
        with pytest.raises(RequestError):
            await json.arrinsert(
                glide_client,
                key,
                "$..a",
                10,  # Index out of range for some paths but valid for others
                ['"out_of_range_value"'],
            )

        with pytest.raises(RequestError):
            await json.arrinsert(
                glide_client,
                key,
                "..a",
                10,  # Index out of range for some paths but valid for others
                ['"out_of_range_value"'],
            )

        # Negative index insertion (should insert from the end of the array)
        result = await json.arrinsert(
            glide_client,
            key,
            "$..a",
            -1,
            ['"negative_index_value"'],
        )
        assert result == [9, 13, None, 11, 10, None, None, 12]  # Update valid paths

        # Check document after negative index insertion
        updated_doc_negative = await json.get(glide_client, key)
        expected_doc["a"].insert(-1, "negative_index_value")
        expected_doc["b"]["a"].insert(-1, "negative_index_value")
        expected_doc["d"][0]["a"].insert(-1, "negative_index_value")
        expected_doc["d"][1]["a"].insert(-1, "negative_index_value")
        expected_doc["f"]["a"].insert(-1, "negative_index_value")
        assert OuterJson.loads(updated_doc_negative) == expected_doc

        # Non-existing path
        with pytest.raises(RequestError):
            await json.arrinsert(glide_client, key, ".path", 5, ['"value"'])

        await json.arrinsert(glide_client, key, "$.path", 5, ['"value"']) == []

        # Key doesnt exist
        with pytest.raises(RequestError):
            await json.arrinsert(glide_client, "non_existent_key", "$", 5, ['"value"'])

        with pytest.raises(RequestError):
            await json.arrinsert(glide_client, "non_existent_key", ".", 5, ['"value"'])

        # value at path is not an array
        with pytest.raises(RequestError):
            await json.arrinsert(glide_client, key, ".e", 5, ['"value"'])

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_debug_fields(self, glide_client: TGlideClient):
        key = get_random_string(10)

        json_value = {
            "key1": 1,
            "key2": 3.5,
            "key3": {"nested_key": {"key1": [4, 5]}},
            "key4": [1, 2, 3],
            "key5": 0,
            "key6": "hello",
            "key7": None,
            "key8": {"nested_key": {"key1": 3.5953862697246314e307}},
            "key9": 3.5953862697246314e307,
            "key10": True,
        }

        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        # Test JSONPath - Fields Subcommand
        # Test integer
        result = await json.debug_fields(glide_client, key, "$.key1")
        assert result == [1]

        # Test float
        result = await json.debug_fields(glide_client, key, "$.key2")
        assert result == [1]

        # Test Nested Value
        result = await json.debug_fields(glide_client, key, "$.key3")
        assert result == [4]

        result = await json.debug_fields(glide_client, key, "$.key3.nested_key.key1")
        assert result == [2]

        # Test Array
        result = await json.debug_fields(glide_client, key, "$.key4[2]")
        assert result == [1]

        # Test String
        result = await json.debug_fields(glide_client, key, "$.key6")
        assert result == [1]

        # Test Null
        result = await json.debug_fields(glide_client, key, "$.key7")
        assert result == [1]

        # Test Bool
        result = await json.debug_fields(glide_client, key, "$.key10")
        assert result == [1]

        # Test all keys
        result = await json.debug_fields(glide_client, key, "$[*]")
        assert result == [1, 1, 4, 3, 1, 1, 1, 2, 1, 1]

        # Test multiple paths
        result = await json.debug_fields(glide_client, key, "$..key1")
        assert result == [1, 2, 1]

        # Test for non-existent path
        result = await json.debug_fields(glide_client, key, "$.key11")
        assert result == []

        # Test for non-existent key
        result = await json.debug_fields(glide_client, "non_existent_key", "$.key10")
        assert result is None

        # Test no provided path
        # Total Fields (19) - breakdown:
        # Top-Level Fields: 10
        # Fields within key3: 4 ($.key3, $.key3.nested_key, $.key3.nested_key.key1, $.key3.nested_key.key1)
        # Fields within key4: 3 ($.key4[0], $.key4[1], $.key4[2])
        # Fields within key8: 2 ($.key8, $.key8.nested_key)
        result = await json.debug_fields(glide_client, key)
        assert result == 19

        # Test legacy path - Fields Subcommand
        # Test integer
        result = await json.debug_fields(glide_client, key, ".key1")
        assert result == 1

        # Test float
        result = await json.debug_fields(glide_client, key, ".key2")
        assert result == 1

        # Test Nested Value
        result = await json.debug_fields(glide_client, key, ".key3")
        assert result == 4

        result = await json.debug_fields(glide_client, key, ".key3.nested_key.key1")
        assert result == 2

        # Test Array
        result = await json.debug_fields(glide_client, key, ".key4[2]")
        assert result == 1

        # Test String
        result = await json.debug_fields(glide_client, key, ".key6")
        assert result == 1

        # Test Null
        result = await json.debug_fields(glide_client, key, ".key7")
        assert result == 1

        # Test Bool
        result = await json.debug_fields(glide_client, key, ".key10")
        assert result == 1

        # Test multiple paths
        result = await json.debug_fields(glide_client, key, "..key1")
        assert result == 1  # Returns number of fields of the first JSON value

        # Test for non-existent path
        with pytest.raises(RequestError):
            await json.debug_fields(glide_client, key, ".key11")

        # Test for non-existent key
        result = await json.debug_fields(glide_client, "non_existent_key", ".key10")
        assert result is None

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_debug_memory(self, glide_client: TGlideClient):
        key = get_random_string(10)

        json_value = {
            "key1": 1,
            "key2": 3.5,
            "key3": {"nested_key": {"key1": [4, 5]}},
            "key4": [1, 2, 3],
            "key5": 0,
            "key6": "hello",
            "key7": None,
            "key8": {"nested_key": {"key1": 3.5953862697246314e307}},
            "key9": 3.5953862697246314e307,
            "key10": True,
        }

        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK
        # Test JSONPath - Memory Subcommand
        # Test integer
        result = await json.debug_memory(glide_client, key, "$.key1")
        assert result == [16]
        # Test float
        result = await json.debug_memory(glide_client, key, "$.key2")
        assert result == [16]
        # Test Nested Value
        result = await json.debug_memory(glide_client, key, "$.key3.nested_key.key1[0]")
        assert result == [16]
        # Test Array
        result = await json.debug_memory(glide_client, key, "$.key4")
        assert result == [16 * 4]

        result = await json.debug_memory(glide_client, key, "$.key4[2]")
        assert result == [16]
        # Test String
        result = await json.debug_memory(glide_client, key, "$.key6")
        assert result == [16]
        # Test Null
        result = await json.debug_memory(glide_client, key, "$.key7")
        assert result == [16]
        # Test Bool
        result = await json.debug_memory(glide_client, key, "$.key10")
        assert result == [16]
        # Test all keys
        result = await json.debug_memory(glide_client, key, "$[*]")
        assert result == [16, 16, 110, 64, 16, 16, 16, 101, 39, 16]
        # Test multiple paths
        result = await json.debug_memory(glide_client, key, "$..key1")
        assert result == [16, 48, 39]
        # Test for non-existent path
        result = await json.debug_memory(glide_client, key, "$.key11")
        assert result == []
        # Test for non-existent key
        result = await json.debug_memory(glide_client, "non_existent_key", "$.key10")
        assert result is None
        # Test no provided path
        # Total Memory (504 bytes) - visual breakdown:
        # ├── Root Object Overhead (129 bytes)
        # └── JSON Elements (374 bytes)
        #    ├── key1: 16 bytes
        #    ├── key2: 16 bytes
        #    ├── key3: 110 bytes
        #    ├── key4: 64 bytes
        #    ├── key5: 16 bytes
        #    ├── key6: 16 bytes
        #    ├── key7: 16 bytes
        #    ├── key8: 101 bytes
        #    └── key9: 39 bytes
        result = await json.debug_memory(glide_client, key)
        assert result == 504
        # Test Legacy Path - Memory Subcommand
        # Test integer
        result = await json.debug_memory(glide_client, key, ".key1")
        assert result == 16
        # Test float
        result = await json.debug_memory(glide_client, key, ".key2")
        assert result == 16
        # Test Nested Value
        result = await json.debug_memory(glide_client, key, ".key3.nested_key.key1[0]")
        assert result == 16
        # Test Array
        result = await json.debug_memory(glide_client, key, ".key4[2]")
        assert result == 16
        # Test String
        result = await json.debug_memory(glide_client, key, ".key6")
        assert result == 16
        # Test Null
        result = await json.debug_memory(glide_client, key, ".key7")
        assert result == 16
        # Test Bool
        result = await json.debug_memory(glide_client, key, ".key10")
        assert result == 16
        # Test multiple paths
        result = await json.debug_memory(glide_client, key, "..key1")
        assert result == 16  # Returns the memory usage of the first JSON value
        # Test for non-existent path
        with pytest.raises(RequestError):
            await json.debug_memory(glide_client, key, ".key11")
        # Test for non-existent key
        result = await json.debug_memory(glide_client, "non_existent_key", ".key10")
        assert result is None

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @typing.no_type_check
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_arrtrim(self, glide_client: TGlideClient):
        key = get_random_string(5)

        # Test with enhanced path syntax
        json_value = '{"a": [0, 1, 2, 3, 4, 5, 6, 7, 8], "b": {"a": [0, 9, 10, 11, 12, 13], "c": {"a": 42}}}'
        assert await json.set(glide_client, key, "$", json_value) == OK

        # Basic trim
        assert await json.arrtrim(glide_client, key, "$..a", 1, 7) == [7, 5, None]
        assert OuterJson.loads(await json.get(glide_client, key, "$..a")) == [
            [1, 2, 3, 4, 5, 6, 7],
            [9, 10, 11, 12, 13],
            42,
        ]

        # Test negative start (should be treated as 0)
        assert await json.arrtrim(glide_client, key, "$.a", -1, 5) == [6]
        assert OuterJson.loads(await json.get(glide_client, key, "$.a")) == [
            [1, 2, 3, 4, 5, 6]
        ]
        assert await json.arrtrim(glide_client, key, ".a", -1, 5) == 6
        assert OuterJson.loads(await json.get(glide_client, key, ".a")) == [
            1,
            2,
            3,
            4,
            5,
            6,
        ]

        # Test end >= size (should be treated as size-1)
        assert await json.arrtrim(glide_client, key, "$.a", 0, 10) == [6]
        assert OuterJson.loads(await json.get(glide_client, key, "$.a")) == [
            [1, 2, 3, 4, 5, 6]
        ]

        assert await json.arrtrim(glide_client, key, ".a", 0, 10) == 6
        assert OuterJson.loads(await json.get(glide_client, key, ".a")) == [
            1,
            2,
            3,
            4,
            5,
            6,
        ]

        # Test start >= size (should empty the array)
        assert await json.arrtrim(glide_client, key, "$.a", 7, 10) == [0]
        assert OuterJson.loads(await json.get(glide_client, key, "$.a")) == [[]]

        assert await json.set(glide_client, key, ".a", '["a", "b", "c"]') == OK
        assert await json.arrtrim(glide_client, key, ".a", 7, 10) == 0
        assert OuterJson.loads(await json.get(glide_client, key, ".a")) == []

        # Test start > end (should empty the array)
        assert await json.arrtrim(glide_client, key, "$..a", 2, 1) == [0, 0, None]
        assert OuterJson.loads(await json.get(glide_client, key, "$..a")) == [
            [],
            [],
            42,
        ]
        assert await json.set(glide_client, key, "..a", '["a", "b", "c", "d"]') == OK
        assert await json.arrtrim(glide_client, key, "..a", 2, 1) == 0
        assert OuterJson.loads(await json.get(glide_client, key, ".a")) == []

        # Multiple path match
        assert await json.set(glide_client, key, "$", json_value) == OK
        assert await json.arrtrim(glide_client, key, "..a", 1, 10) == 8
        assert OuterJson.loads(await json.get(glide_client, key, "$..a")) == [
            [1, 2, 3, 4, 5, 6, 7, 8],
            [9, 10, 11, 12, 13],
            42,
        ]

        # Test with non-existent path
        with pytest.raises(RequestError):
            await json.arrtrim(glide_client, key, ".non_existent", 0, 1)

        assert await json.arrtrim(glide_client, key, "$.non_existent", 0, 1) == []

        # Test with non-array path
        assert await json.arrtrim(glide_client, key, "$", 0, 1) == [None]

        with pytest.raises(RequestError):
            await json.arrtrim(glide_client, key, ".", 0, 1)

        # Test with non-existent key
        with pytest.raises(RequestError):
            await json.arrtrim(glide_client, "non_existent_key", "$", 0, 1)

        # Test with non-existent key
        with pytest.raises(RequestError):
            await json.arrtrim(glide_client, "non_existent_key", ".", 0, 1)

        # Test empty array
        assert await json.set(glide_client, key, "$.empty", "[]") == OK
        assert await json.arrtrim(glide_client, key, "$.empty", 0, 1) == [0]
        assert await json.arrtrim(glide_client, key, ".empty", 0, 1) == 0
        assert OuterJson.loads(await json.get(glide_client, key, "$.empty")) == [[]]

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_arrindex(self, glide_client: TGlideClient):
        key = get_random_string(10)

        json_value = {
            "empty_array": [],
            "single_element": ["apple"],
            "multiple_elements": ["banana", "cherry", "date"],
            "nested_arrays": [
                ["alpha"],
                ["beta", "gamma"],
                ["delta", "epsilon", "zeta"],
            ],
            "mixed_types": [1, "two", True, None, 5.5],
            "not_array": 5,
            "nested_arrays2": [
                ["a"],
                ["ab", "abc"],
                ["abcd", "abcde", "abcdef", "abcdefg", "abcdefgh", 1, 2, None, "gamma"],
            ],
            "nested_structure": {
                "level1": {
                    "level2": {
                        "level3": [
                            ["gamma", "theta"],
                            ["iota", "kappa", "gamma"],
                        ]
                    }
                }
            },
        }

        assert await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == OK

        # JSONPath Syntax Tests
        # Search for "beta" in all arrays at the root level, Non-array values return null
        result = await json.arrindex(glide_client, key, "$[*]", '"beta"')
        assert result == [-1, -1, -1, -1, -1, None, -1, None]

        # Search for a boolean
        result = await json.arrindex(glide_client, key, "$.mixed_types", "true")
        assert result == [2]  # True found at index 2 in the "mixed_types" array

        # Search for a float
        result = await json.arrindex(glide_client, key, "$.mixed_types", "5.5")
        assert result == [4]  # 5.5 found at index 4 in the "mixed_types" array

        # Search for "gamma" at nested level
        result = await json.arrindex(glide_client, key, "$.nested_arrays[*]", '"gamma"')
        assert result == [-1, 1, -1]  # "gamma" found at index 1 in the second array

        # Search for "gamma" at nested level with a specified range
        result = await json.arrindex(
            glide_client,
            key,
            "$.nested_arrays2[*]",
            '"gamma"',
            JsonArrIndexOptions(start=0, end=5),
        )
        assert result == [-1, -1, -1]

        # Search for "gamma" at nested level with start > end
        result = await json.arrindex(
            glide_client,
            key,
            "$.nested_arrays[*]",
            '"gamma"',
            JsonArrIndexOptions(start=2, end=1),
        )
        assert result == [-1, -1, -1]  # Invalid range, returns -1 for all

        # Search for "omega" which does not exist
        result = await json.arrindex(glide_client, key, "$[*]", '"omega"')
        assert result == [-1, -1, -1, -1, -1, None, -1, None]  # "omega" not found

        # Search for null values, null found at at third index in the fifth array
        result = await json.arrindex(glide_client, key, "$[*]", "null")
        assert result == [-1, -1, -1, -1, 3, None, -1, None]

        # Search in mixed types, "two" found at first index in the fifth array
        result = await json.arrindex(glide_client, key, "$[*]", '"two"')
        assert result == [-1, -1, -1, -1, 1, None, -1, None]

        # Out of range check for "start" value
        result = await json.arrindex(
            glide_client, key, "$[*]", '"apple"', JsonArrIndexOptions(start=-200)
        )
        assert result == [
            -1,
            0,
            -1,
            -1,
            -1,
            None,
            -1,
            None,
        ]  # Rounded to the array's start

        # Check for end = -1, tests if the function includes the last element, found "gamma" at index 8 at the third array
        result = await json.arrindex(
            glide_client,
            key,
            "$.nested_arrays2[*]",
            '"gamma"',
            JsonArrIndexOptions(start=0, end=-1),
        )
        assert result == [-1, -1, 8]

        # Check for non-existent key
        with pytest.raises(RequestError):
            await json.arrindex(
                glide_client,
                "Non_existent",
                "$.nested_arrays2[*]",
                '"abcdefg"',
                JsonArrIndexOptions(start=0, end=-1),
            )

        # Check for non-existent path
        result = await json.arrindex(
            glide_client,
            key,
            "$.nested_arrays3[*]",
            '"abcdefg"',
            JsonArrIndexOptions(start=0, end=-1),
        )
        assert result == []

        # Using JSONPath syntax to search for "gamma" in nested_structure.level1.level2.level3
        result = await json.arrindex(
            glide_client, key, "$.nested_structure.level1.level2.level3[*]", '"gamma"'
        )
        assert result == [
            0,
            2,
        ]  # "gamma" at index 0 in first array, index 2 in second array

        # Check for inclusive behavior of start in JSONPath syntax
        result = await json.arrindex(
            glide_client,
            key,
            "$.nested_structure.level1.level2.level3[*]",
            '"gamma"',
            JsonArrIndexOptions(start=0),
        )
        assert result == [
            0,
            2,
        ]  # "gamma" at index 0 of level3[0] and index 2 of level3[1].

        # Check for exclusive behavior of end in JSONPath syntax
        result = await json.arrindex(
            glide_client,
            key,
            "$.nested_structure.level1.level2.level3[*]",
            '"gamma"',
            JsonArrIndexOptions(start=0, end=2),
        )
        assert result == [
            0,
            -1,
        ]  # Only "gamma" at index 0 of level3[0] is found; gamma at index 2 of level3[1] is excluded as its not within the
        # search range.

        # Check for passing start = 0, end = 0 in JSONPath syntax
        result = await json.arrindex(
            glide_client,
            key,
            "$.nested_arrays[2]",
            '"zeta"',
            JsonArrIndexOptions(start=0, end=0),
        )
        assert result == [2]  # "zeta" found at index 2 as the whole range was searched

        # Check for passing start = 1, end = 0 (start>end) but end is a "special value" in JSONPath syntax
        result = await json.arrindex(
            glide_client,
            key,
            "$.nested_arrays[2]",
            '"zeta"',
            JsonArrIndexOptions(start=1, end=0),
        )
        assert result == [2]  # "zeta" found at index 2 as the whole range was searched

        # Check for passing start = 1, end = -1 (start>end) but end is a "special value" in JSONPath syntax
        result = await json.arrindex(
            glide_client,
            key,
            "$.nested_arrays[2]",
            '"zeta"',
            JsonArrIndexOptions(start=1, end=-1),
        )
        assert result == [2]  # "zeta" found at index 2 as the whole range was searched

        # Restricted Path Syntax Tests
        # Search for "abcd" in the "nested_arrays2" array
        result = await json.arrindex(glide_client, key, ".nested_arrays2[2]", '"abcd"')
        assert result == 0  # "abcd" found at index 0

        # Search for "abcd" in the "nested_arrays2" array with specified range
        result = await json.arrindex(
            glide_client,
            key,
            ".nested_arrays2[2]",
            '"abcd"',
            JsonArrIndexOptions(start=1, end=4),
        )
        assert result == -1  # "abcd" not found at the specified range

        # Search for "abcdefg" in the "nested_arrays2" with start > end
        result = await json.arrindex(
            glide_client,
            key,
            ".nested_arrays2[2]",
            '"abcdefg"',
            JsonArrIndexOptions(start=4, end=3),
        )
        assert result == -1

        # Search for "theta" which does not exist
        result = await json.arrindex(glide_client, key, ".multiple_elements", '"theta"')
        assert result == -1  # "theta" not found

        # Check for non_existent path
        with pytest.raises(RequestError):
            await json.arrindex(glide_client, key, ".non_existent", '"value"')

        # Search in an empty array
        result = await json.arrindex(glide_client, key, ".empty_array", '"anything"')
        assert result == -1  # Nothing to find in empty array

        # Search for a boolean
        result = await json.arrindex(glide_client, key, ".mixed_types", "true")
        assert result == 2  # True found at index 2

        # Search for a float
        result = await json.arrindex(glide_client, key, ".mixed_types", "5.5")
        assert result == 4  # 5.5 found at index 4

        # Search for null value
        result = await json.arrindex(glide_client, key, ".mixed_types", "null")
        assert result == 3  # null found at index 3

        # Out of range check for "start" value
        result = await json.arrindex(
            glide_client,
            key,
            ".single_element",
            '"apple"',
            JsonArrIndexOptions(start=-200),
        )
        assert result == 0  # Rounded to the array's start

        # Check for end = -1, tests if the function includes the last element
        result = await json.arrindex(
            glide_client,
            key,
            ".nested_arrays2[2]",
            '"gamma"',
            JsonArrIndexOptions(start=0, end=-1),
        )
        assert result == 8

        # Check for non-existent key
        with pytest.raises(RequestError):
            await json.arrindex(
                glide_client, "Non_existent", ".nested_arrays2[1]", '"abcdefg"'
            )

        # Check for value at path is not an array
        with pytest.raises(RequestError):
            await json.arrindex(glide_client, key, ".not_array", "val")

        # Using legacy syntax to search for "gamma" in nested_structure
        result = await json.arrindex(
            glide_client, key, ".nested_structure.level1.level2.level3[*]", '"gamma"'
        )
        assert result == 0  # Legacy syntax returns index from first matching array

        # Check for inclusive behavior of start in legacy syntax
        result = await json.arrindex(
            glide_client,
            key,
            ".nested_arrays[2]",
            '"epsilon"',
            JsonArrIndexOptions(start=1),
        )
        assert result == 1  # "epsilon" found at index 1 in nested_arrays[2].

        # Check for exclusive behavior of end in legacy syntax
        result = await json.arrindex(
            glide_client,
            key,
            ".nested_arrays[2]",
            '"zeta"',
            JsonArrIndexOptions(start=1, end=2),
        )
        assert result == -1  # "zeta" at index 2 is excluded due to exclusive end.

        # Check for passing start = 0, end = 0
        result = await json.arrindex(
            glide_client,
            key,
            ".nested_arrays[2]",
            '"zeta"',
            JsonArrIndexOptions(start=0, end=0),
        )
        assert result == 2  # "zeta" found at index 2 as the whole range was searched

        # Check for passing start = 1, end = 0 (start>end) but end is a "special value"
        result = await json.arrindex(
            glide_client,
            key,
            ".nested_arrays[2]",
            '"zeta"',
            JsonArrIndexOptions(start=1, end=0),
        )
        assert result == 2  # "zeta" found at index 2 as the whole range was searched

        # Check for passing start = 1, end = -1 (start>end) but end is a "special value"
        result = await json.arrindex(
            glide_client,
            key,
            ".nested_arrays[2]",
            '"zeta"',
            JsonArrIndexOptions(start=1, end=-1),
        )
        assert result == 2  # "zeta" found at index 2 as the whole range was searched

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_arrappend(self, glide_client: TGlideClient):
        key = get_random_string(10)
        initial_json_value = '{"a": 1, "b": ["one", "two"]}'
        assert await json.set(glide_client, key, "$", initial_json_value) == OK

        assert await json.arrappend(glide_client, key, "$.b", ['"three"']) == [3]
        assert await json.arrappend(glide_client, key, ".b", ['"four"', '"five"']) == 5

        result = await json.get(glide_client, key, "$")
        assert isinstance(result, bytes)
        assert OuterJson.loads(result) == [
            {"a": 1, "b": ["one", "two", "three", "four", "five"]}
        ]

        assert await json.arrappend(glide_client, key, "$.a", ['"value"']) == [None]

        # JSONPath, path doesnt exist
        assert await json.arrappend(glide_client, key, "$.c", ['"value"']) == []
        # Legacy path, `path` doesnt exist
        with pytest.raises(RequestError):
            await json.arrappend(glide_client, key, ".c", ['"value"'])

        # Legacy path, the JSON value at `path` is not a array
        with pytest.raises(RequestError):
            await json.arrappend(glide_client, key, ".a", ['"value"'])

        with pytest.raises(RequestError):
            await json.arrappend(glide_client, "non_existing_key", "$.b", ['"six"'])
        with pytest.raises(RequestError):
            await json.arrappend(glide_client, "non_existing_key", ".b", ['"six"'])

        # multiple path match
        json_value = '[[], ["a"], ["a", "b"]]'
        assert await json.set(glide_client, key, "$", json_value) == OK
        assert await json.arrappend(glide_client, key, "[*]", ['"c"']) == 1
        result = await json.get(glide_client, key, "$")
        assert isinstance(result, bytes)
        assert OuterJson.loads(result) == [[["c"], ["a", "c"], ["a", "b", "c"]]]

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_resp(self, glide_client: TGlideClient):
        key = get_random_string(5)

        # Generate random JSON content with specified types
        json_value = {
            "obj": {"a": get_random_value("int"), "b": get_random_value("float")},
            "arr": [get_random_value("int") for _ in range(3)],
            "str": get_random_value("str"),
            "bool": get_random_value("bool"),
            "int": get_random_value("int"),
            "float": get_random_value("float"),
            "nullVal": get_random_value("null"),
        }

        json_value_expected = copy.deepcopy(json_value)
        json_value_expected["obj"]["b"] = str(json_value["obj"]["b"]).encode()
        json_value_expected["float"] = str(json_value["float"]).encode()
        json_value_expected["str"] = str(json_value["str"]).encode()
        json_value_expected["bool"] = str(json_value["bool"]).lower().encode()
        assert (
            await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == "OK"
        )

        assert await json.resp(glide_client, key, "$.*") == [
            [
                b"{",
                [b"a", json_value_expected["obj"]["a"]],
                [b"b", json_value_expected["obj"]["b"]],
            ],
            [b"[", *json_value_expected["arr"]],
            json_value_expected["str"],
            json_value_expected["bool"],
            json_value_expected["int"],
            json_value_expected["float"],
            json_value_expected["nullVal"],
        ]

        # multiple path match, the first will be returned
        assert await json.resp(glide_client, key, "*") == [
            b"{",
            [b"a", json_value_expected["obj"]["a"]],
            [b"b", json_value_expected["obj"]["b"]],
        ]

        assert await json.resp(glide_client, key, "$") == [
            [
                b"{",
                [
                    b"obj",
                    [
                        b"{",
                        [b"a", json_value_expected["obj"]["a"]],
                        [b"b", json_value_expected["obj"]["b"]],
                    ],
                ],
                [b"arr", [b"[", *json_value_expected["arr"]]],
                [
                    b"str",
                    json_value_expected["str"],
                ],
                [
                    b"bool",
                    json_value_expected["bool"],
                ],
                [b"int", json_value["int"]],
                [
                    b"float",
                    json_value_expected["float"],
                ],
                [b"nullVal", json_value["nullVal"]],
            ],
        ]

        assert await json.resp(glide_client, key, "$.str") == [
            json_value_expected["str"]
        ]
        assert await json.resp(glide_client, key, ".str") == json_value_expected["str"]

        # Further tests with a new random JSON structure
        json_value = {
            "a": [random.randint(1, 10) for _ in range(3)],
            "b": {
                "a": [random.randint(1, 10) for _ in range(2)],
                "c": {"a": random.randint(1, 10)},
            },
        }
        assert (
            await json.set(glide_client, key, "$", OuterJson.dumps(json_value)) == "OK"
        )

        # Multiple path match
        assert await json.resp(glide_client, key, "$..a") == [
            [b"[", *json_value["a"]],
            [b"[", *json_value["b"]["a"]],
            json_value["b"]["c"]["a"],
        ]

        assert await json.resp(glide_client, key, "..a") == [b"[", *json_value["a"]]

        # Test for non-existent paths
        assert await json.resp(glide_client, key, "$.nonexistent") == []
        with pytest.raises(RequestError):
            await json.resp(glide_client, key, "nonexistent")

        # Test for non-existent key
        assert await json.resp(glide_client, "nonexistent_key", "$") is None
        assert await json.resp(glide_client, "nonexistent_key", ".") is None
        assert await json.resp(glide_client, "nonexistent_key") is None

    @pytest.mark.parametrize("cluster_mode", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_arrpop(self, glide_client: TGlideClient):
        key = get_random_string(5)
        key2 = get_random_string(5)

        json_value = '{"a": [1, 2, true], "b": {"a": [3, 4, ["value", 3, false] ,5], "c": {"a": 42}}}'
        assert await json.set(glide_client, key, "$", json_value) == OK

        assert await json.arrpop(
            glide_client, key, JsonArrPopOptions(path="$.a", index=1)
        ) == [b"2"]
        assert (
            await json.arrpop(glide_client, key, JsonArrPopOptions(path="$..a"))
        ) == [b"true", b"5", None]

        assert (
            await json.arrpop(glide_client, key, JsonArrPopOptions(path="..a")) == b"1"
        )
        # Even if only one array element was returned, ensure second array at `..a` was popped
        assert await json.get(glide_client, key, "$..a") == b"[[],[3,4],42]"

        # Out of index
        assert await json.arrpop(
            glide_client, key, JsonArrPopOptions(path="$..a", index=10)
        ) == [None, b"4", None]

        assert (
            await json.arrpop(
                glide_client, key, JsonArrPopOptions(path="..a", index=-10)
            )
            == b"3"
        )

        # Path is not an array
        assert await json.arrpop(glide_client, key, JsonArrPopOptions(path="$")) == [
            None
        ]
        with pytest.raises(RequestError):
            assert await json.arrpop(glide_client, key, JsonArrPopOptions(path="."))
        with pytest.raises(RequestError):
            assert await json.arrpop(glide_client, key)

        # Non existing path
        assert (
            await json.arrpop(
                glide_client, key, JsonArrPopOptions(path="$.non_existing_path")
            )
            == []
        )
        with pytest.raises(RequestError):
            assert await json.arrpop(
                glide_client, key, JsonArrPopOptions(path="non_existing_path")
            )

        # Non existing key
        with pytest.raises(RequestError):
            await json.arrpop(
                glide_client, "non_existing_key", JsonArrPopOptions(path="$.a")
            )
        with pytest.raises(RequestError):
            await json.arrpop(
                glide_client, "non_existing_key", JsonArrPopOptions(path=".a")
            )

        assert (
            await json.set(glide_client, key2, "$", '[[], ["a"], ["a", "b", "c"]]')
            == OK
        )
        assert (
            await json.arrpop(glide_client, key2, JsonArrPopOptions(path=".", index=-1))
            == b'["a","b","c"]'
        )
        assert await json.arrpop(glide_client, key2) == b'["a"]'

        # pop from an empty array
        assert await json.arrpop(glide_client, key2, JsonArrPopOptions("$[0]")) == [
            None
        ]
        assert await json.arrpop(glide_client, key2, JsonArrPopOptions("$[0]", 10)) == [
            None
        ]
        assert await json.arrpop(glide_client, key2, JsonArrPopOptions("[0]")) is None
        assert (
            await json.arrpop(glide_client, key2, JsonArrPopOptions("[0]", 10)) is None
        )

        # non jsonpath pops from all matching paths, even if one result is being returned
        assert (
            await json.set(
                glide_client, key2, "$", '[[], ["a"], ["a", "b"], ["a", "b", "c"]]'
            )
            == OK
        )

        assert await json.arrpop(glide_client, key2, JsonArrPopOptions("[*]")) == b'"a"'
        assert await json.get(glide_client, key2, ".") == b'[[],[],["a"],["a","b"]]'

    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("is_atomic", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_batch_array(
        self, glide_client: GlideClusterClient, is_atomic: bool
    ):
        transaction = ClusterBatch(is_atomic=is_atomic)

        key = get_random_string(5)
        json_value1 = {"a": 1.0, "b": 2}
        json_value2 = {"a": 1.0, "b": [1, 2]}

        # Test 'set', 'get', and 'clear' commands
        json_batch.set(transaction, key, "$", OuterJson.dumps(json_value1))
        json_batch.clear(transaction, key, "$")
        json_batch.set(transaction, key, "$", OuterJson.dumps(json_value1))
        json_batch.get(transaction, key, ".")

        # Test array related commands
        json_batch.set(transaction, key, "$", OuterJson.dumps(json_value2))
        json_batch.arrappend(transaction, key, "$.b", ["3", "4"])
        json_batch.arrindex(transaction, key, "$.b", "2")
        json_batch.arrinsert(transaction, key, "$.b", 2, ["5"])
        json_batch.arrlen(transaction, key, "$.b")
        json_batch.arrpop(transaction, key, JsonArrPopOptions(path="$.b", index=2))
        json_batch.arrtrim(transaction, key, "$.b", 1, 2)
        json_batch.get(transaction, key, ".")

        result = await glide_client.exec(transaction, raise_on_error=False)
        assert isinstance(result, list)

        assert result[0] == "OK"  # set
        assert result[1] == 1  # clear
        assert result[2] == "OK"  # set
        assert isinstance(result[3], bytes)
        assert OuterJson.loads(result[3]) == json_value1  # get

        assert result[4] == "OK"  # set
        assert result[5] == [4]  # arrappend
        assert result[6] == [1]  # arrindex
        assert result[7] == [5]  # arrinsert
        assert result[8] == [5]  # arrlen
        assert result[9] == [b"5"]  # arrpop
        assert result[10] == [2]  # arrtrim
        assert isinstance(result[11], bytes)
        assert OuterJson.loads(result[11]) == {"a": 1.0, "b": [2, 3]}  # get

    @pytest.mark.parametrize("cluster_mode", [True])
    @pytest.mark.parametrize("is_atomic", [True, False])
    @pytest.mark.parametrize("protocol", [ProtocolVersion.RESP2, ProtocolVersion.RESP3])
    async def test_json_batch(self, glide_client: GlideClusterClient, is_atomic: bool):
        transaction = ClusterBatch(is_atomic=is_atomic)

        key = f"{{key}}-1{get_random_string(5)}"
        key2 = f"{{key}}-2{get_random_string(5)}"
        key3 = f"{{key}}-3{get_random_string(5)}"
        json_value = {"a": [1, 2], "b": [3, 4], "c": "c", "d": True}

        json_batch.set(transaction, key, "$", OuterJson.dumps(json_value))

        # Test debug commands
        json_batch.debug_memory(transaction, key, "$.a")
        json_batch.debug_fields(transaction, key, "$.a")

        # Test obj commands
        json_batch.objlen(transaction, key, ".")
        json_batch.objkeys(transaction, key, ".")

        # Test num commands
        json_batch.numincrby(transaction, key, "$.a[*]", 10.0)
        json_batch.nummultby(transaction, key, "$.a[*]", 10.0)

        # Test str commands
        json_batch.strappend(transaction, key, '"-test"', "$.c")
        json_batch.strlen(transaction, key, "$.c")

        # Test type command
        json_batch.type(transaction, key, "$.a")

        # Test mget command
        json_value2 = {"b": [3, 4], "c": "c", "d": True}
        json_batch.set(transaction, key2, "$", OuterJson.dumps(json_value2))
        json_batch.mget(transaction, [key, key2, key3], "$.a")

        # Test toggle command
        json_batch.toggle(transaction, key, "$.d")

        # Test resp command
        json_batch.resp(transaction, key, "$")

        # Test del command
        json_batch.delete(transaction, key, "$.d")

        # Test forget command
        json_batch.forget(transaction, key, "$.c")

        result = await glide_client.exec(transaction, raise_on_error=False)
        assert isinstance(result, list)

        assert result[0] == "OK"  # set
        assert result[1] == [48]  # debug_memory
        assert result[2] == [2]  # debug_field

        assert result[3] == 4  # objlen
        assert result[4] == [b"a", b"b", b"c", b"d"]  # objkeys
        assert result[5] == b"[11,12]"  # numincrby
        assert result[6] == b"[110,120]"  # nummultby
        assert result[7] == [6]  # strappend
        assert result[8] == [6]  # strlen
        assert result[9] == [b"array"]  # type
        assert result[10] == "OK"  # set
        assert result[11] == [b"[[110,120]]", b"[]", None]  # mget
        assert result[12] == [False]  # toggle

        assert result[13] == [
            [
                b"{",
                [b"a", [b"[", 110, 120]],
                [b"b", [b"[", 3, 4]],
                [b"c", b"c-test"],
                [b"d", b"false"],
            ]
        ]  # resp

        assert result[14] == 1  # del
        assert result[15] == 1  # forget



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0
"""Glide module for `JSON` commands in batch.

Examples:
    >>> import json
    >>> from glide import json_batch
    >>> batch = ClusterBatch(is_atomic=True)
    >>> value = {'a': 1.0, 'b': 2}
    >>> json_str = json.dumps(value) # Convert Python dictionary to JSON string using json.dumps()
    >>> json_batch.set(batch, "doc", "$", json_str)
    >>> json_batch.get(batch, "doc", "$") # Returns the value at path '$' in the JSON document stored at `doc` as
                                                # JSON string.
    >>> result = await glide_client.exec(batch)
    >>> print result[0] # set result
        'OK'  # Indicates successful setting of the value at path '$' in the key stored at `doc`.
    >>> print result[1] # get result
        b"[{\"a\":1.0,\"b\":2}]"
    >>> print json.loads(str(result[1]))
        [{"a": 1.0, "b": 2}] # JSON object retrieved from the key `doc` using json.loads()

"""

from typing import List, Optional, Union

from glide.async_commands.batch import TBatch
from glide.async_commands.core import ConditionalChange
from glide.async_commands.server_modules.glide_json import (
    JsonArrIndexOptions,
    JsonArrPopOptions,
    JsonGetOptions,
)
from glide.constants import TEncodable


def set(
    batch: TBatch,
    key: TEncodable,
    path: TEncodable,
    value: TEncodable,
    set_condition: Optional[ConditionalChange] = None,
) -> TBatch:
    """
    Sets the JSON value at the specified `path` stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (TEncodable): Represents the path within the JSON document where the value will be set.
            The key will be modified only if `value` is added as the last child in the specified `path`, or if the specified
            `path` acts as the parent of a new child being added.
        value (TEncodable): The value to set at the specific path, in JSON formatted bytes or str.
        set_condition (Optional[ConditionalChange]): Set the value only if the given condition is met (within the key or path).
            Equivalent to [`XX` | `NX`] in the RESP API. Defaults to None.

    Command response:
        Optional[TOK]: If the value is successfully set, returns OK.
            If `value` isn't set because of `set_condition`, returns None.
    """
    args = ["JSON.SET", key, path, value]
    if set_condition:
        args.append(set_condition.value)

    return batch.custom_command(args)


def get(
    batch: TBatch,
    key: TEncodable,
    paths: Optional[Union[TEncodable, List[TEncodable]]] = None,
    options: Optional[JsonGetOptions] = None,
) -> TBatch:
    """
    Retrieves the JSON value at the specified `paths` stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        paths (Optional[Union[TEncodable, List[TEncodable]]]): The path or list of paths within the JSON document.
            Default to None.
        options (Optional[JsonGetOptions]): Options for formatting the byte representation of the JSON data.
            See `JsonGetOptions`.

    Command response:
        TJsonResponse[Optional[bytes]]:
            If one path is given:
                For JSONPath (path starts with `$`):
                    Returns a stringified JSON list of bytes replies for every possible path,
                    or a byte string representation of an empty array, if path doesn't exists.
                    If `key` doesn't exist, returns None.
                For legacy path (path doesn't start with `$`):
                    Returns a byte string representation of the value in `path`.
                    If `path` doesn't exist, an error is raised.
                    If `key` doesn't exist, returns None.
            If multiple paths are given:
                Returns a stringified JSON object in bytes, in which each path is a key, and it's corresponding value, is the
                value as if the path was executed in the command as a single path.
        In case of multiple paths, and `paths` are a mix of both JSONPath and legacy path, the command behaves as if all are
        JSONPath paths.
        For more information about the returned type, see `TJsonResponse`.
    """
    args = ["JSON.GET", key]
    if options:
        args.extend(options.get_options())
    if paths:
        if isinstance(paths, (str, bytes)):
            paths = [paths]
        args.extend(paths)

    return batch.custom_command(args)


def mget(
    batch: TBatch,
    keys: List[TEncodable],
    path: TEncodable,
) -> TBatch:
    """
    Retrieves the JSON values at the specified `path` stored at multiple `keys`.

    Note:
        When in cluster mode:
        - If the batch is an atomic batch (transaction), then all keys must map to the same slot.
        - If the batch is a non-atomic batch (pipeline), and the keys map to different hash slots,
            the command will be split across these slots and executed separately for each.
            This means the command is atomic only at the slot level. If one or more slot-specific
            requests fail, the entire call will return the first encountered error, even
            though some requests may have succeeded while others did not.
            If this behavior impacts your application logic, consider splitting the
            request into sub-requests per slot to ensure atomicity.

    Args:
        batch (TBatch): The batch to execute the command.
        keys (List[TEncodable]): A list of keys for the JSON documents.
        path (TEncodable): The path within the JSON documents.

    Command response:
        List[Optional[bytes]]:
            For JSONPath (`path` starts with `$`):
                Returns a list of byte representations of the values found at the given path for each key.
                If `path` does not exist within the key, the entry will be an empty array.
            For legacy path (`path` doesn't starts with `$`):
                Returns a list of byte representations of the values found at the given path for each key.
                If `path` does not exist within the key, the entry will be None.
            If a key doesn't exist, the corresponding list element will be None.
    """
    args = ["JSON.MGET"] + keys + [path]
    return batch.custom_command(args)


def arrappend(
    batch: TBatch,
    key: TEncodable,
    path: TEncodable,
    values: List[TEncodable],
) -> TBatch:
    """
    Appends one or more `values` to the JSON array at the specified `path` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (TEncodable): Represents the path within the JSON document where the `values` will be appended.
        values (TEncodable): The values to append to the JSON array at the specified path.
            JSON string values must be wrapped with quotes. For example, to append `"foo"`, pass `"\"foo\""`.

    Command response:
        TJsonResponse[int]:
            For JSONPath (`path` starts with `$`):
                Returns a list of integer replies for every possible path, indicating the new length of the array after
                appending `values`, or None for JSON values matching the path that are not an array.
                If `path` doesn't exist, an empty array will be returned.
            For legacy path (`path` doesn't start with `$`):
                Returns the length of the array after appending `values` to the array at `path`.
                If multiple paths match, the length of the first updated array is returned.
                If the JSON value at `path` is not a array or if `path` doesn't exist, an error is raised.
            If `key` doesn't exist, an error is raised.
        For more information about the returned type, see `TJsonResponse`.
    """
    args = ["JSON.ARRAPPEND", key, path] + values
    return batch.custom_command(args)


def arrindex(
    batch: TBatch,
    key: TEncodable,
    path: TEncodable,
    value: TEncodable,
    options: Optional[JsonArrIndexOptions] = None,
) -> TBatch:
    """
    Searches for the first occurrence of a scalar JSON value (i.e., a value that is neither an object nor an array) within
    arrays at the specified `path` in the JSON document stored at `key`.

    If specified, `options.start` and `options.end` define an inclusive-to-exclusive search range within the array.
    (Where `options.start` is inclusive and `options.end` is exclusive).

    Out-of-range indices adjust to the nearest valid position, and negative values count from the end
    (e.g., `-1` is the last element, `-2` the second last).

    Setting `options.end` to `0` behaves like `-1`, extending the range to the array's end (inclusive).

    If `options.start` exceeds `options.end`, `-1` is returned, indicating that the value was not found.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (TEncodable): The path within the JSON document.
        value (TEncodable): The value to search for within the arrays.
        options (Optional[JsonArrIndexOptions]): Options specifying an inclusive `start` index and an optional exclusive `end`
            index for a range-limited search.
            Defaults to the full array if not provided. See `JsonArrIndexOptions`.

    Command response:
        Optional[Union[int, List[int]]]:
            For JSONPath (`path` starts with `$`):
                Returns an array of integers for every possible path, indicating of the first occurrence of `value` within the
                array, or None for JSON values matching the path that are not an array.
                A returned value of `-1` indicates that the value was not found in that particular array.
                If `path` does not exist, an empty array will be returned.
            For legacy path (`path` doesn't start with `$`):
                Returns an integer representing the index of the first occurrence of `value` within the array at the specified
                path.
                A returned value of `-1` indicates that the value was not found in that particular array.
                If multiple paths match, the index of the value from the first matching array is returned.
                If the JSON value at the `path` is not an array or if `path` does not exist, an error is raised.
            If `key` does not exist, an error is raised.
    """
    args = ["JSON.ARRINDEX", key, path, value]

    if options:
        args.extend(options.to_args())

    return batch.custom_command(args)


def arrinsert(
    batch: TBatch,
    key: TEncodable,
    path: TEncodable,
    index: int,
    values: List[TEncodable],
) -> TBatch:
    """
    Inserts one or more values into the array at the specified `path` within the JSON document stored at `key`, before the
    given `index`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (TEncodable): The path within the JSON document.
        index (int): The array index before which values are inserted.
        values (List[TEncodable]): The JSON values to be inserted into the array, in JSON formatted bytes or str.
            Json string values must be wrapped with single quotes. For example, to append "foo", pass '"foo"'.

    Command response:
        TJsonResponse[int]:
            For JSONPath (`path` starts with '$'):
                Returns a list of integer replies for every possible path, indicating the new length of the array,
                or None for JSON values matching the path that are not an array.
                If `path` does not exist, an empty array will be returned.
            For legacy path (`path` doesn't start with '$'):
                Returns an integer representing the new length of the array.
                If multiple paths are matched, returns the length of the first modified array.
                If `path` doesn't exist or the value at `path` is not an array, an error is raised.
            If the index is out of bounds, an error is raised.
            If `key` doesn't exist, an error is raised.
    """
    args = ["JSON.ARRINSERT", key, path, str(index)] + values
    return batch.custom_command(args)


def arrlen(
    batch: TBatch,
    key: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Retrieves the length of the array at the specified `path` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[TEncodable]): The path within the JSON document. Defaults to None.

    Command response:
        Optional[TJsonResponse[int]]:
            For JSONPath (`path` starts with `$`):
                Returns a list of integer replies for every possible path, indicating the length of the array,
                or None for JSON values matching the path that are not an array.
                If `path` doesn't exist, an empty array will be returned.
            For legacy path (`path` doesn't starts with `$`):
                Returns the length of the array at `path`.
                If multiple paths match, the length of the first array match is returned.
                If the JSON value at `path` is not a array or if `path` doesn't exist, an error is raised.
            If `key` doesn't exist, None is returned.
    """
    args = ["JSON.ARRLEN", key]
    if path:
        args.append(path)
    return batch.custom_command(args)


def arrpop(
    batch: TBatch,
    key: TEncodable,
    options: Optional[JsonArrPopOptions] = None,
) -> TBatch:
    """
    Pops an element from the array located at the specified path within the JSON document stored at `key`.
    If `options.index` is provided, it pops the element at that index instead of the last element.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        options (Optional[JsonArrPopOptions]): Options including the path and optional index. See `JsonArrPopOptions`.
            Default to None.
            If not specified, attempts to pop the last element from the root value if it's an array.
            If the root value is not an array, an error will be raised.

    Command response:
        Optional[TJsonResponse[bytes]]:
            For JSONPath (`options.path` starts with `$`):
                Returns a list of bytes string replies for every possible path, representing the popped JSON values,
                or None for JSON values matching the path that are not an array or are an empty array.
                If `options.path` doesn't exist, an empty list will be returned.
            For legacy path (`options.path` doesn't starts with `$`):
                Returns a bytes string representing the popped JSON value, or None if the array at `options.path` is empty.
                If multiple paths match, the value from the first matching array that is not empty is returned.
                If the JSON value at `options.path` is not a array or if `options.path` doesn't exist, an error is raised.
            If `key` doesn't exist, an error is raised.
    """
    args = ["JSON.ARRPOP", key]
    if options:
        args.extend(options.to_args())

    return batch.custom_command(args)


def arrtrim(
    batch: TBatch,
    key: TEncodable,
    path: TEncodable,
    start: int,
    end: int,
) -> TBatch:
    """
    Trims an array at the specified `path` within the JSON document stored at `key` so that it becomes a subarray
    [start, end], both inclusive.
    If `start` < 0, it is treated as 0.
    If `end` >= size (size of the array), it is treated as size-1.
    If `start` >= size or `start` > `end`, the array is emptied and 0 is returned.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (TEncodable): The path within the JSON document.
        start (int): The start index, inclusive.
        end (int): The end index, inclusive.

    Command response:
        TJsonResponse[int]:
            For JSONPath (`path` starts with '$'):
                Returns a list of integer replies for every possible path, indicating the new length of the array, or None for
                JSON values matching the path that are not an array.
                If a value is an empty array, its corresponding return value is 0.
                If `path` doesn't exist, an empty array will be returned.
            For legacy path (`path` doesn't starts with `$`):
                Returns an integer representing the new length of the array.
                If the array is empty, returns 0.
                If multiple paths match, the length of the first trimmed array match is returned.
                If `path` doesn't exist, or the value at `path` is not an array, an error is raised.
            If `key` doesn't exist, an error is raised.
    """

    return batch.custom_command(["JSON.ARRTRIM", key, path, str(start), str(end)])


def clear(
    batch: TBatch,
    key: TEncodable,
    path: Optional[str] = None,
) -> TBatch:
    """
    Clears arrays or objects at the specified JSON path in the document stored at `key`.
    Numeric values are set to `0`, and boolean values are set to `False`, and string values are converted to empty strings.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[str]): The path within the JSON document. Default to None.

    Command response:
        int: The number of containers cleared, numeric values zeroed, and booleans toggled to `false`,
        and string values converted to empty strings.
        If `path` doesn't exist, or the value at `path` is already empty (e.g., an empty array, object, or string),
        0 is returned.
        If `key doesn't exist, an error is raised.
    """
    args = ["JSON.CLEAR", key]
    if path:
        args.append(path)

    return batch.custom_command(args)


def debug_fields(
    batch: TBatch,
    key: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Returns the number of fields of the JSON value at the specified `path` within the JSON document stored at `key`.
    - **Primitive Values**: Each non-container JSON value (e.g., strings, numbers, booleans, and null) counts as one field.
    - **Arrays and Objects:**: Each item in an array and each key-value pair in an object is counted as one field.
        (Each top-level value counts as one field, regardless of it's type.)
        - Their nested values are counted recursively and added to the total.
        - **Example**: For the JSON `{"a": 1, "b": [2, 3, {"c": 4}]}`, the count would be:
            - Top-level: 2 fields (`"a"` and `"b"`)
            - Nested: 3 fields in the array (`2`, `3`, and `{"c": 4}`) plus 1 for the object (`"c"`)
            - Total: 2 (top-level) + 3 (from array) + 1 (from nested object) = 6 fields.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[TEncodable]): The path within the JSON document. Defaults to root if not provided.

    Command response:
        Optional[TJsonUniversalResponse[int]]:
            For JSONPath (`path` starts with `$`):
                Returns an array of integers, each indicating the number of fields for each matched `path`.
                If `path` doesn't exist, an empty array will be returned.
            For legacy path (`path` doesn't start with `$`):
                Returns an integer indicating the number of fields for each matched `path`.
                If multiple paths match, number of fields of the first JSON value match is returned.
                If `path` doesn't exist, an error is raised.
            If `path` is not provided, it reports the total number of fields in the entire JSON document.
            If `key` doesn't exist, None is returned.
    """
    args = ["JSON.DEBUG", "FIELDS", key]
    if path:
        args.append(path)

    return batch.custom_command(args)


def debug_memory(
    batch: TBatch,
    key: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Reports memory usage in bytes of a JSON value at the specified `path` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[TEncodable]): The path within the JSON document. Defaults to None.

    Command response:
        Optional[TJsonUniversalResponse[int]]:
            For JSONPath (`path` starts with `$`):
                Returns an array of integers, indicating the memory usage in bytes of a JSON value for each matched `path`.
                If `path` doesn't exist, an empty array will be returned.
            For legacy path (`path` doesn't start with `$`):
                Returns an integer, indicating the memory usage in bytes for the JSON value in `path`.
                If multiple paths match, the memory usage of the first JSON value match is returned.
                If `path` doesn't exist, an error is raised.
            If `path` is not provided, it reports the total memory usage in bytes in the entire JSON document.
            If `key` doesn't exist, None is returned.
    """
    args = ["JSON.DEBUG", "MEMORY", key]
    if path:
        args.append(path)

    return batch.custom_command(args)


def delete(
    batch: TBatch,
    key: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Deletes the JSON value at the specified `path` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[TEncodable]): The path within the JSON document.
            If None, deletes the entire JSON document at `key`. Defaults to None.

    Command response:
        int: The number of elements removed.
        If `key` or `path` doesn't exist, returns 0.
    """

    return batch.custom_command(["JSON.DEL", key] + ([path] if path else []))


def forget(
    batch: TBatch,
    key: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Deletes the JSON value at the specified `path` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[TEncodable]): The path within the JSON document.
            If None, deletes the entire JSON document at `key`. Defaults to None.

    Command response:
        int: The number of elements removed.
        If `key` or `path` doesn't exist, returns 0.
    """

    return batch.custom_command(["JSON.FORGET", key] + ([path] if path else []))


def numincrby(
    batch: TBatch,
    key: TEncodable,
    path: TEncodable,
    number: Union[int, float],
) -> TBatch:
    """
    Increments or decrements the JSON value(s) at the specified `path` by `number` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (TEncodable): The path within the JSON document.
        number (Union[int, float]): The number to increment or decrement by.

    Command response:
        bytes:
            For JSONPath (`path` starts with `$`):
                Returns a bytes string representation of an array of bulk strings, indicating the new values after
                incrementing for each matched `path`.
                If a value is not a number, its corresponding return value will be `null`.
                If `path` doesn't exist, a byte string representation of an empty array will be returned.
            For legacy path (`path` doesn't start with `$`):
                Returns a bytes string representation of the resulting value after the increment or decrement.
                If multiple paths match, the result of the last updated value is returned.
                If the value at the `path` is not a number or `path` doesn't exist, an error is raised.
            If `key` does not exist, an error is raised.
            If the result is out of the range of 64-bit IEEE double, an error is raised.
    """
    args = ["JSON.NUMINCRBY", key, path, str(number)]

    return batch.custom_command(args)


def nummultby(
    batch: TBatch,
    key: TEncodable,
    path: TEncodable,
    number: Union[int, float],
) -> TBatch:
    """
    Multiplies the JSON value(s) at the specified `path` by `number` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (TEncodable): The path within the JSON document.
        number (Union[int, float]): The number to multiply by.

    Command response:
        bytes:
            For JSONPath (`path` starts with `$`):
                Returns a bytes string representation of an array of bulk strings, indicating the new values after
                multiplication for each matched `path`.
                If a value is not a number, its corresponding return value will be `null`.
                If `path` doesn't exist, a byte string representation of an empty array will be returned.
            For legacy path (`path` doesn't start with `$`):
                Returns a bytes string representation of the resulting value after multiplication.
                If multiple paths match, the result of the last updated value is returned.
                If the value at the `path` is not a number or `path` doesn't exist, an error is raised.
            If `key` does not exist, an error is raised.
            If the result is out of the range of 64-bit IEEE double, an error is raised.
    """
    args = ["JSON.NUMMULTBY", key, path, str(number)]

    return batch.custom_command(args)


def objlen(
    batch: TBatch,
    key: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Retrieves the number of key-value pairs in the object stored at the specified `path` within the JSON document stored at
    `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[TEncodable]): The path within the JSON document. Defaults to None.

    Command response:
        Optional[TJsonResponse[int]]:
            For JSONPath (`path` starts with `$`):
                Returns a list of integer replies for every possible path, indicating the length of the object,
                or None for JSON values matching the path that are not an object.
                If `path` doesn't exist, an empty array will be returned.
            For legacy path (`path` doesn't starts with `$`):
                Returns the length of the object at `path`.
                If multiple paths match, the length of the first object match is returned.
                If the JSON value at `path` is not an object or if `path` doesn't exist, an error is raised.
            If `key` doesn't exist, None is returned.
    """
    args = ["JSON.OBJLEN", key]
    if path:
        args.append(path)

    return batch.custom_command(args)


def objkeys(
    batch: TBatch,
    key: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Retrieves key names in the object values at the specified `path` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[TEncodable]): Represents the path within the JSON document where the key names will be retrieved.
            Defaults to None.

    Command response:
        Optional[TJsonUniversalResponse[List[bytes]]]:
            For JSONPath (`path` starts with `$`):
                Returns a list of arrays containing key names for each matching object.
                If a value matching the path is not an object, an empty array is returned.
                If `path` doesn't exist, an empty array is returned.
            For legacy path (`path` starts with `.`):
                Returns a list of key names for the object value matching the path.
                If multiple objects match the path, the key names of the first object are returned.
                If a value matching the path is not an object, an error is raised.
                If `path` doesn't exist, None is returned.
            If `key` doesn't exist, None is returned.
    """
    args = ["JSON.OBJKEYS", key]
    if path:
        args.append(path)

    return batch.custom_command(args)


def resp(
    batch: TBatch,
    key: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Retrieve the JSON value at the specified `path` within the JSON document stored at `key`.
    The returning result is in the Valkey or Redis OSS Serialization Protocol (RESP).\n
    JSON null is mapped to the RESP Null Bulk String.\n
    JSON Booleans are mapped to RESP Simple string.\n
    JSON integers are mapped to RESP Integers.\n
    JSON doubles are mapped to RESP Bulk Strings.\n
    JSON strings are mapped to RESP Bulk Strings.\n
    JSON arrays are represented as RESP arrays, where the first element is the simple string
    [, followed by the array's elements.\n
    JSON objects are represented as RESP object, where the first element is the simple string {, followed by key-value pairs,
    each of which is a RESP bulk string.\n

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[TEncodable]): The path within the JSON document. Default to None.

    Command response:
        TJsonUniversalResponse[Optional[Union[bytes, int, List[Optional[Union[bytes, int]]]]]]
            For JSONPath ('path' starts with '$'):
                Returns a list of replies for every possible path, indicating the RESP form of the JSON value.
                If `path` doesn't exist, returns an empty list.
            For legacy path (`path` doesn't starts with `$`):
                Returns a single reply for the JSON value at the specified path, in its RESP form.
                This can be a bytes object, an integer, None, or a list representing complex structures.
                If multiple paths match, the value of the first JSON value match is returned.
                If `path` doesn't exist, an error is raised.
            If `key` doesn't exist, an None is returned.
    """
    args = ["JSON.RESP", key]
    if path:
        args.append(path)

    return batch.custom_command(args)


def strappend(
    batch: TBatch,
    key: TEncodable,
    value: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Appends the specified `value` to the string stored at the specified `path` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        value (TEncodable): The value to append to the string. Must be wrapped with single quotes. For example,
            to append "foo", pass '"foo"'.
        path (Optional[TEncodable]): The path within the JSON document. Default to None.

    Command response:
        TJsonResponse[int]:
            For JSONPath (`path` starts with `$`):
                Returns a list of integer replies for every possible path, indicating the length of the resulting string after
                appending `value`,
                or None for JSON values matching the path that are not string.
                If `key` doesn't exist, an error is raised.
            For legacy path (`path` doesn't start with `$`):
                Returns the length of the resulting string after appending `value` to the string at `path`.
                If multiple paths match, the length of the last updated string is returned.
                If the JSON value at `path` is not a string of if `path` doesn't exist, an error is raised.
                If `key` doesn't exist, an error is raised.
        For more information about the returned type, see `TJsonResponse`.
    """
    return batch.custom_command(
        ["JSON.STRAPPEND", key] + ([path, value] if path else [value])
    )


def strlen(
    batch: TBatch,
    key: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Returns the length of the JSON string value stored at the specified `path` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[TEncodable]): The path within the JSON document. Default to None.

    Command response:
        TJsonResponse[Optional[int]]:
            For JSONPath (`path` starts with `$`):
                Returns a list of integer replies for every possible path, indicating the length of the JSON string value,
                or None for JSON values matching the path that are not string.
            For legacy path (`path` doesn't start with `$`):
                Returns the length of the JSON value at `path` or None if `key` doesn't exist.
                If multiple paths match, the length of the first mached string is returned.
                If the JSON value at `path` is not a string of if `path` doesn't exist, an error is raised.
            If `key` doesn't exist, None is returned.
        For more information about the returned type, see `TJsonResponse`.
    """
    return batch.custom_command(
        ["JSON.STRLEN", key, path] if path else ["JSON.STRLEN", key]
    )


def toggle(
    batch: TBatch,
    key: TEncodable,
    path: TEncodable,
) -> TBatch:
    """
    Toggles a Boolean value stored at the specified `path` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (TEncodable): The path within the JSON document. Default to None.

    Command response:
        TJsonResponse[bool]:
            For JSONPath (`path` starts with `$`):
                Returns a list of boolean replies for every possible path, with the toggled boolean value,
                or None for JSON values matching the path that are not boolean.
                If `key` doesn't exist, an error is raised.
            For legacy path (`path` doesn't start with `$`):
                Returns the value of the toggled boolean in `path`.
                If the JSON value at `path` is not a boolean of if `path` doesn't exist, an error is raised.
                If `key` doesn't exist, an error is raised.
        For more information about the returned type, see `TJsonResponse`.
    """
    return batch.custom_command(["JSON.TOGGLE", key, path])


def type(
    batch: TBatch,
    key: TEncodable,
    path: Optional[TEncodable] = None,
) -> TBatch:
    """
    Retrieves the type of the JSON value at the specified `path` within the JSON document stored at `key`.

    Args:
        batch (TBatch): The batch to execute the command.
        key (TEncodable): The key of the JSON document.
        path (Optional[TEncodable]): The path within the JSON document. Default to None.

    Command response:
        Optional[TJsonUniversalResponse[bytes]]:
            For JSONPath ('path' starts with '$'):
                Returns a list of byte string replies for every possible path, indicating the type of the JSON value.
                If `path` doesn't exist, an empty array will be returned.
            For legacy path (`path` doesn't starts with `$`):
                Returns the type of the JSON value at `path`.
                If multiple paths match, the type of the first JSON value match is returned.
                If `path` doesn't exist, None will be returned.
            If `key` doesn't exist, None is returned.
    """
    args = ["JSON.TYPE", key]
    if path:
        args.append(path)

    return batch.custom_command(args)



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

import random
from typing import AsyncGenerator, List, Optional, Union

import pytest
from glide.config import (
    AdvancedGlideClientConfiguration,
    AdvancedGlideClusterClientConfiguration,
    BackoffStrategy,
    GlideClientConfiguration,
    GlideClusterClientConfiguration,
    NodeAddress,
    ProtocolVersion,
    ReadFrom,
    ServerCredentials,
)
from glide.exceptions import ClosingError
from glide.glide_client import GlideClient, GlideClusterClient, TGlideClient
from glide.logger import Level as logLevel
from glide.logger import Logger
from glide.routes import AllNodes

from tests.utils.cluster import ValkeyCluster
from tests.utils.utils import (
    check_if_server_version_lt,
    set_new_acl_username_with_password,
)

DEFAULT_HOST = "localhost"
DEFAULT_PORT = 6379
DEFAULT_TEST_LOG_LEVEL = logLevel.OFF

Logger.set_logger_config(DEFAULT_TEST_LOG_LEVEL)


def pytest_addoption(parser):
    parser.addoption(
        "--host",
        default=DEFAULT_HOST,
        action="store",
        help="Server host endpoint, defaults to `%(default)s`",
    )

    parser.addoption(
        "--port",
        default=DEFAULT_PORT,
        action="store",
        help="Server port, defaults to `%(default)s`",
    )

    parser.addoption(
        "--tls",
        default=False,
        action="store_true",
        help="TLS enabled, defaults to `%(default)s`",
    )

    parser.addoption(
        "--no-tls",
        dest="tls",
        action="store_false",
        help="TLS disabled, defaults to `%(default)s`",
    )

    parser.addoption(
        "--load-module",
        action="append",
        help="""Load additional Valkey modules (provide full path for the module's shared library).
            Use multiple times for multiple modules.
            Example:
            pytest -v --load-module=/path/to/module1.so --load-module=/path/to/module2.so""",
        default=[],
    )

    parser.addoption(
        "--cluster-endpoints",
        action="store",
        help="""Comma-separated list of cluster endpoints for standalone cluster in the format host1:port1,host2:port2,...
            Note: The cluster will be flashed between tests.
            Example:
                pytest -v --asyncio-mode=auto --cluster-endpoints=127.0.0.1:6379
                pytest -v --asyncio-mode=auto --cluster-endpoints=127.0.0.1:6379,127.0.0.1:6380
            """,
        default=None,
    )

    parser.addoption(
        "--standalone-endpoints",
        action="store",
        help="""Comma-separated list of cluster endpoints for cluster mode cluster in the format host1:port1,host2:port2,...
            Note: The cluster will be flashed between tests.
            Example:
                pytest -v --asyncio-mode=auto --standalone-endpoints=127.0.0.1:6379
                pytest -v --asyncio-mode=auto --standalone-endpoints=127.0.0.1:6379,127.0.0.1:6380
            """,
        default=None,
    )


def parse_endpoints(endpoints_str: str) -> List[List[str]]:
    """
    Parse the endpoints string into a list of lists containing host and port.
    """
    try:
        endpoints = [endpoint.split(":") for endpoint in endpoints_str.split(",")]
        for endpoint in endpoints:
            if len(endpoint) != 2:
                raise ValueError(
                    "Each endpoint should be in the format 'host:port'.\nEndpoints should be separated by commas."
                )
            host, port = endpoint
            if not host or not port.isdigit():
                raise ValueError(
                    "Both host and port should be specified and port should be a valid integer."
                )
        return endpoints
    except ValueError as e:
        raise ValueError("Invalid endpoints format: " + str(e))


def create_clusters(tls, load_module, cluster_endpoints, standalone_endpoints):
    """
    Create Valkey clusters based on the provided options.
    """
    if cluster_endpoints or standalone_endpoints:
        # Endpoints were passed by the caller, not creating clusters internally
        if cluster_endpoints:
            cluster_endpoints = parse_endpoints(cluster_endpoints)
            pytest.valkey_cluster = ValkeyCluster(tls=tls, addresses=cluster_endpoints)
        if standalone_endpoints:
            standalone_endpoints = parse_endpoints(standalone_endpoints)
            pytest.standalone_cluster = ValkeyCluster(
                tls=tls, addresses=standalone_endpoints
            )
    else:
        # No endpoints were provided, create new clusters
        pytest.valkey_cluster = ValkeyCluster(
            tls=tls,
            cluster_mode=True,
            load_module=load_module,
            addresses=cluster_endpoints,
            replica_count=2,
        )
        pytest.standalone_cluster = ValkeyCluster(
            tls=tls,
            cluster_mode=False,
            shard_count=1,
            replica_count=1,
            load_module=load_module,
            addresses=standalone_endpoints,
        )


@pytest.fixture(autouse=True, scope="session")
def call_before_all_pytests(request):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    tls = request.config.getoption("--tls")
    load_module = request.config.getoption("--load-module")
    cluster_endpoints = request.config.getoption("--cluster-endpoints")
    standalone_endpoints = request.config.getoption("--standalone-endpoints")

    create_clusters(tls, load_module, cluster_endpoints, standalone_endpoints)


def pytest_sessionfinish(session, exitstatus):
    """
    Called after whole test run finished, right before
    returning the exit status to the system.
    """
    try:
        del pytest.valkey_cluster
    except AttributeError:
        # valkey_cluster was not set, skip deletion
        pass

    try:
        del pytest.standalone_cluster
    except AttributeError:
        # standalone_cluster was not set, skip deletion
        pass


def pytest_collection_modifyitems(config, items):
    """
    Modify collected test items.

    This function checks if cluster or standalone endpoints are provided. If so, it checks if the test requires
    cluster mode and skips it accordingly.
    """
    for item in items:
        if config.getoption("--cluster-endpoints") or config.getoption(
            "--standalone-endpoints"
        ):
            if "cluster_mode" in item.fixturenames:
                cluster_mode_value = item.callspec.params.get("cluster_mode", None)
                if cluster_mode_value is True and not config.getoption(
                    "--cluster-endpoints"
                ):
                    item.add_marker(
                        pytest.mark.skip(
                            reason="Test skipped because cluster_mode=True and cluster endpoints weren't provided"
                        )
                    )
                elif cluster_mode_value is False and not config.getoption(
                    "--standalone-endpoints"
                ):
                    item.add_marker(
                        pytest.mark.skip(
                            reason="Test skipped because cluster_mode=False and standalone endpoints weren't provided"
                        )
                    )


@pytest.fixture(scope="function")
async def glide_client(
    request,
    cluster_mode: bool,
    protocol: ProtocolVersion,
) -> AsyncGenerator[TGlideClient, None]:
    "Get async socket client for tests"
    client = await create_client(
        request, cluster_mode, protocol=protocol, request_timeout=5000
    )
    yield client
    await test_teardown(request, cluster_mode, protocol)
    await client.close()


@pytest.fixture(scope="function")
async def management_client(
    request,
    cluster_mode: bool,
    protocol: ProtocolVersion,
) -> AsyncGenerator[TGlideClient, None]:
    "Get async socket client for tests, used to manage the state when tests are on the client ability to connect"
    client = await create_client(request, cluster_mode, protocol=protocol)
    yield client
    await test_teardown(request, cluster_mode, protocol)
    await client.close()


@pytest.fixture(scope="function")
async def acl_glide_client(
    request,
    cluster_mode: bool,
    protocol: ProtocolVersion,
    management_client: TGlideClient,
) -> AsyncGenerator[TGlideClient, None]:
    """
    Client fot tests that use a server pre-configured with an ACL user.
    This function first uses the management client to register the USERNAME with INITIAL_PASSWORD,so that
    the client would be ablt to connect.
    It then returns a client with this USERNAME and INITIAL_PASSWORD already set as its ServerCredentials.
    """

    await set_new_acl_username_with_password(
        management_client, USERNAME, INITIAL_PASSWORD
    )

    client = await create_client(
        request,
        cluster_mode,
        protocol=protocol,
        credentials=ServerCredentials(username=USERNAME, password=INITIAL_PASSWORD),
    )
    yield client
    await test_teardown(request, cluster_mode, protocol)
    await client.close()


async def create_client(
    request,
    cluster_mode: bool,
    credentials: Optional[ServerCredentials] = None,
    database_id: int = 0,
    addresses: Optional[List[NodeAddress]] = None,
    client_name: Optional[str] = None,
    protocol: ProtocolVersion = ProtocolVersion.RESP3,
    request_timeout: Optional[int] = 1000,
    connection_timeout: Optional[int] = 1000,
    cluster_mode_pubsub: Optional[
        GlideClusterClientConfiguration.PubSubSubscriptions
    ] = None,
    standalone_mode_pubsub: Optional[
        GlideClientConfiguration.PubSubSubscriptions
    ] = None,
    inflight_requests_limit: Optional[int] = None,
    read_from: ReadFrom = ReadFrom.PRIMARY,
    client_az: Optional[str] = None,
    reconnect_strategy: Optional[BackoffStrategy] = None,
    valkey_cluster: Optional[ValkeyCluster] = None,
) -> Union[GlideClient, GlideClusterClient]:
    # Create async socket client
    use_tls = request.config.getoption("--tls")
    if cluster_mode:
        valkey_cluster = valkey_cluster or pytest.valkey_cluster  # type: ignore
        assert type(valkey_cluster) is ValkeyCluster
        assert database_id == 0
        k = min(3, len(valkey_cluster.nodes_addr))
        seed_nodes = random.sample(valkey_cluster.nodes_addr, k=k)
        cluster_config = GlideClusterClientConfiguration(
            addresses=seed_nodes if addresses is None else addresses,
            use_tls=use_tls,
            credentials=credentials,
            client_name=client_name,
            protocol=protocol,
            request_timeout=request_timeout,
            pubsub_subscriptions=cluster_mode_pubsub,
            inflight_requests_limit=inflight_requests_limit,
            read_from=read_from,
            client_az=client_az,
            advanced_config=AdvancedGlideClusterClientConfiguration(connection_timeout),
        )
        return await GlideClusterClient.create(cluster_config)
    else:
        assert type(pytest.standalone_cluster) is ValkeyCluster  # type: ignore
        config = GlideClientConfiguration(
            addresses=(
                pytest.standalone_cluster.nodes_addr if addresses is None else addresses  # type: ignore
            ),
            use_tls=use_tls,
            credentials=credentials,
            database_id=database_id,
            client_name=client_name,
            protocol=protocol,
            request_timeout=request_timeout,
            pubsub_subscriptions=standalone_mode_pubsub,
            inflight_requests_limit=inflight_requests_limit,
            read_from=read_from,
            client_az=client_az,
            advanced_config=AdvancedGlideClientConfiguration(connection_timeout),
            reconnect_strategy=reconnect_strategy,
        )
        return await GlideClient.create(config)


USERNAME = "username"
INITIAL_PASSWORD = "initial_password"
NEW_PASSWORD = "new_secure_password"
WRONG_PASSWORD = "wrong_password"


async def auth_client(client: TGlideClient, password: str, username: str = "default"):
    """
    Authenticates the given TGlideClient server connected. If no username is provided, uses the 'default' user.
    """
    if isinstance(client, GlideClient):
        return await client.custom_command(["AUTH", username, password])
    elif isinstance(client, GlideClusterClient):
        return await client.custom_command(
            ["AUTH", username, password], route=AllNodes()
        )


async def config_set_new_password(client: TGlideClient, password: str):
    """
    Sets a new password for the given TGlideClient server connected.
    This function updates the server to require a new password.
    """
    if isinstance(client, GlideClient):
        await client.config_set({"requirepass": password})
    elif isinstance(client, GlideClusterClient):
        await client.config_set({"requirepass": password}, route=AllNodes())


async def kill_connections(client: TGlideClient):
    """
    Kills all connections to the given TGlideClient server connected.
    """
    if isinstance(client, GlideClient):
        await client.custom_command(["CLIENT", "KILL", "TYPE", "normal"])
    elif isinstance(client, GlideClusterClient):
        await client.custom_command(
            ["CLIENT", "KILL", "TYPE", "normal"], route=AllNodes()
        )


async def test_teardown(request, cluster_mode: bool, protocol: ProtocolVersion):
    """
    Perform teardown tasks such as flushing all data from the cluster.

    If authentication is required, attempt to connect with the known password,
    reset it back to empty, and proceed with teardown.
    """
    credentials = None
    try:
        # Try connecting without credentials
        client = await create_client(
            request, cluster_mode, protocol=protocol, request_timeout=2000
        )
        await client.custom_command(["FLUSHALL"])
        await client.close()
    except ClosingError as e:
        # Check if the error is due to authentication
        if "NOAUTH" in str(e):
            # Use the known password to authenticate
            credentials = ServerCredentials(password=NEW_PASSWORD)
            client = await create_client(
                request,
                cluster_mode,
                protocol=protocol,
                request_timeout=2000,
                credentials=credentials,
            )
            try:
                await auth_client(client, NEW_PASSWORD)
                # Reset the server password back to empty
                await config_set_new_password(client, "")
                await client.update_connection_password(None)
                # Perform the teardown
                await client.custom_command(["FLUSHALL"])
            finally:
                await client.close()
        else:
            raise e


@pytest.fixture(autouse=True)
async def skip_if_version_below(request):
    """
    Skip test(s) if server version is below than given parameter. Can skip a complete test suite.

    Example:
        @pytest.mark.skip_if_version_below('7.0.0')
        async def test_meow_meow(...):
            ...
    """
    if request.node.get_closest_marker("skip_if_version_below"):
        min_version = request.node.get_closest_marker("skip_if_version_below").args[0]
        client = await create_client(request, False)
        if await check_if_server_version_lt(client, min_version):
            pytest.skip(
                reason=f"This feature added in version {min_version}",
                allow_module_level=True,
            )



# Copyright Valkey GLIDE Project Contributors - SPDX Identifier: Apache-2.0

from enum import Enum
from typing import List, Optional, Tuple, Union, cast

from glide.async_commands.command_args import Limit, OrderBy
from glide.constants import TEncodable


class InfBound(Enum):
    """
    Enumeration representing numeric and lexicographic positive and negative infinity bounds for sorted set.
    """

    POS_INF = {"score_arg": "+inf", "lex_arg": "+"}
    """
    Positive infinity bound for sorted set.

    `score_arg`: represents numeric positive infinity (+inf).

    `lex_arg`: represents lexicographic positive infinity (+).
    """
    NEG_INF = {"score_arg": "-inf", "lex_arg": "-"}
    """
    Negative infinity bound for sorted set.

    `score_arg`: represents numeric negative infinity (-inf).

    `lex_arg`: represents lexicographic negative infinity (-).
    """


class AggregationType(Enum):
    """
    Enumeration representing aggregation types for `ZINTERSTORE` and `ZUNIONSTORE` sorted set commands.
    """

    SUM = "SUM"
    """
    Represents aggregation by summing the scores of elements across inputs where they exist.
    """
    MIN = "MIN"
    """
    Represents aggregation by selecting the minimum score of an element across inputs where it exists.
    """
    MAX = "MAX"
    """
    Represents aggregation by selecting the maximum score of an element across inputs where it exists.
    """


class ScoreFilter(Enum):
    """
    Defines which elements to pop from a sorted set.

    ScoreFilter is a mandatory option for [ZMPOP](https://valkey.io/commands/zmpop)
    and [BZMPOP](https://valkey.io/commands/bzmpop).
    """

    MIN = "MIN"
    """
    Pop elements with the lowest scores.
    """
    MAX = "MAX"
    """
    Pop elements with the highest scores.
    """


class ScoreBoundary:
    """
    Represents a specific numeric score boundary in a sorted set.

    Args:
        value (float): The score value.
        is_inclusive (bool): Whether the score value is inclusive. Defaults to True.
    """

    def __init__(self, value: float, is_inclusive: bool = True):
        # Convert the score boundary to Valkey protocol format
        self.value = str(value) if is_inclusive else f"({value}"


class LexBoundary:
    """
    Represents a specific lexicographic boundary in a sorted set.

    Args:
        value (str): The lex value.
        is_inclusive (bool): Whether the score value is inclusive. Defaults to True.
    """

    def __init__(self, value: str, is_inclusive: bool = True):
        # Convert the lexicographic boundary to Valkey protocol format
        self.value = f"[{value}" if is_inclusive else f"({value}"


class RangeByIndex:
    """
    Represents a range by index (rank) in a sorted set.

    The `start` and `end` arguments represent zero-based indexes.

    Attributes:
        start (int): The start index of the range.
        end (int): The end index of the range.
    """

    def __init__(self, start: int, end: int):
        self.start = start
        self.end = end


class RangeByScore:
    """
    Represents a range by score in a sorted set.

    The `start` and `end` arguments represent score boundaries.

    Attributes:
        start (Union[InfBound, ScoreBoundary]): The start score boundary.
        end (Union[InfBound, ScoreBoundary]): The end score boundary.
        limit (Optional[Limit]): The limit argument for a range query. Defaults to None. See `Limit`
            class for more information.
    """

    def __init__(
        self,
        start: Union[InfBound, ScoreBoundary],
        end: Union[InfBound, ScoreBoundary],
        limit: Optional[Limit] = None,
    ):
        self.start = (
            start.value["score_arg"] if isinstance(start, InfBound) else start.value
        )
        self.end = end.value["score_arg"] if isinstance(end, InfBound) else end.value
        self.limit = limit


class RangeByLex:
    """
    Represents a range by lexicographical order in a sorted set.

    The `start` and `end` arguments represent lexicographical boundaries.

    Attributes:
        start (Union[InfBound, LexBoundary]): The start lexicographic boundary.
        end (Union[InfBound, LexBoundary]): The end lexicographic boundary.
        limit (Optional[Limit]): The limit argument for a range query. Defaults to None. See `Limit` class
            for more information.
    """

    def __init__(
        self,
        start: Union[InfBound, LexBoundary],
        end: Union[InfBound, LexBoundary],
        limit: Optional[Limit] = None,
    ):
        self.start = (
            start.value["lex_arg"] if isinstance(start, InfBound) else start.value
        )
        self.end = end.value["lex_arg"] if isinstance(end, InfBound) else end.value
        self.limit = limit


class GeospatialData:
    """
    Represents a geographic position defined by longitude and latitude.

    The exact limits, as specified by EPSG:900913 / EPSG:3785 / OSGEO:41001 are the following:

        - Valid longitudes are from -180 to 180 degrees.
        - Valid latitudes are from -85.05112878 to 85.05112878 degrees.

    Attributes:
        longitude (float): The longitude coordinate.
        latitude (float): The latitude coordinate.
    """

    def __init__(self, longitude: float, latitude: float):
        self.longitude = longitude
        self.latitude = latitude


class GeoUnit(Enum):
    """
    Enumeration representing distance units options for the `GEODIST` command.
    """

    METERS = "m"
    """
    Represents distance in meters.
    """
    KILOMETERS = "km"
    """
    Represents distance in kilometers.
    """
    MILES = "mi"
    """
    Represents distance in miles.
    """
    FEET = "ft"
    """
    Represents distance in feet.
    """


class GeoSearchByRadius:
    """
    Represents search criteria of searching within a certain radius from a specified point.

    Attributes:
        radius (float): Radius of the search area.
        unit (GeoUnit): Unit of the radius. See `GeoUnit`.
    """

    def __init__(self, radius: float, unit: GeoUnit):
        """
        Initialize the search criteria.
        """
        self.radius = radius
        self.unit = unit

    def to_args(self) -> List[str]:
        """
        Convert the search criteria to the corresponding part of the command.

        Returns:
            List[str]: List representation of the search criteria.
        """
        return ["BYRADIUS", str(self.radius), self.unit.value]


class GeoSearchByBox:
    """
    Represents search criteria of searching within a specified rectangular area.

    Attributes:
        width (float): Width of the bounding box.
        height (float): Height of the bounding box
        unit (GeoUnit): Unit of the radius. See `GeoUnit`.
    """

    def __init__(self, width: float, height: float, unit: GeoUnit):
        """
        Initialize the search criteria.
        """
        self.width = width
        self.height = height
        self.unit = unit

    def to_args(self) -> List[str]:
        """
        Convert the search criteria to the corresponding part of the command.

        Returns:
            List[str]: List representation of the search criteria.
        """
        return ["BYBOX", str(self.width), str(self.height), self.unit.value]


class GeoSearchCount:
    """
    Represents the count option for limiting the number of results in a GeoSearch.

    Attributes:
        count (int): The maximum number of results to return.
        any_option (bool): Whether to allow returning as enough matches are found.
            This means that the results returned may not be the ones closest to the specified point. Default to False.
    """

    def __init__(self, count: int, any_option: bool = False):
        """
        Initialize the count option.
        """
        self.count = count
        self.any_option = any_option

    def to_args(self) -> List[str]:
        """
        Convert the count option to the corresponding part of the command.

        Returns:
            List[str]: List representation of the count option.
        """
        if self.any_option:
            return ["COUNT", str(self.count), "ANY"]
        return ["COUNT", str(self.count)]


def _create_zrange_args(
    key: TEncodable,
    range_query: Union[RangeByLex, RangeByScore, RangeByIndex],
    reverse: bool,
    with_scores: bool,
    destination: Optional[TEncodable] = None,
) -> List[TEncodable]:
    args = [destination] if destination else []
    args += [key, str(range_query.start), str(range_query.end)]

    if isinstance(range_query, RangeByScore):
        args.append("BYSCORE")
    elif isinstance(range_query, RangeByLex):
        args.append("BYLEX")
    if reverse:
        args.append("REV")
    if hasattr(range_query, "limit") and range_query.limit is not None:
        args.extend(
            [
                "LIMIT",
                str(range_query.limit.offset),
                str(range_query.limit.count),
            ]
        )
    if with_scores:
        args.append("WITHSCORES")

    return args


def separate_keys(
    keys: Union[List[TEncodable], List[Tuple[TEncodable, float]]],
) -> Tuple[List[TEncodable], List[TEncodable]]:
    """
    Returns separate lists of keys and weights in case of weighted keys.
    """
    if not keys:
        return [], []

    key_list: List[TEncodable] = []
    weight_list: List[TEncodable] = []

    if isinstance(keys[0], tuple):
        for item in keys:
            key = item[0]
            weight = item[1]
            key_list.append(cast(TEncodable, key))
            weight_list.append(cast(TEncodable, str(weight)))
    else:
        key_list.extend(cast(List[TEncodable], keys))

    return key_list, weight_list


def _create_zinter_zunion_cmd_args(
    keys: Union[List[TEncodable], List[Tuple[TEncodable, float]]],
    aggregation_type: Optional[AggregationType] = None,
    destination: Optional[TEncodable] = None,
) -> List[TEncodable]:
    args: List[TEncodable] = []

    if destination:
        args.append(destination)

    args.append(str(len(keys)))

    only_keys, weights = separate_keys(keys)

    args.extend(only_keys)

    if weights:
        args.append("WEIGHTS")
        args.extend(weights)

    if aggregation_type:
        args.append("AGGREGATE")
        args.append(aggregation_type.value)

    return args


def _create_geosearch_args(
    keys: List[TEncodable],
    search_from: Union[str, bytes, GeospatialData],
    search_by: Union[GeoSearchByRadius, GeoSearchByBox],
    order_by: Optional[OrderBy] = None,
    count: Optional[GeoSearchCount] = None,
    with_coord: bool = False,
    with_dist: bool = False,
    with_hash: bool = False,
    store_dist: bool = False,
) -> List[TEncodable]:
    args: List[TEncodable] = keys
    if isinstance(search_from, (str, bytes)):
        args.extend(["FROMMEMBER", search_from])
    else:
        args.extend(
            [
                "FROMLONLAT",
                str(search_from.longitude),
                str(search_from.latitude),
            ]
        )

    args.extend(search_by.to_args())

    if order_by:
        args.append(order_by.value)
    if count:
        args.extend(count.to_args())

    if with_coord:
        args.append("WITHCOORD")
    if with_dist:
        args.append("WITHDIST")
    if with_hash:
        args.append("WITHHASH")

    if store_dist:
        args.append("STOREDIST")

    return args
